# sage.doctest: needs sage.combinat sage.modules
r"""
Littelmann paths

AUTHORS:

- Mark Shimozono, Anne Schilling (2012): Initial version
- Anne Schilling (2013): Implemented
  :class:`~sage.combinat.crystals.littelmann_path.CrystalOfProjectedLevelZeroLSPaths`
- Travis Scrimshaw (2016): Implemented
  :class:`~sage.combinat.crystals.littelmann_path.InfinityCrystalOfLSPaths`
"""
# ***************************************************************************
#       Copyright (C) 2012 Mark Shimozono
#                          Anne Schilling
#                     2016 Travis Scrimshaw <tcscrims at gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ***************************************************************************

from sage.misc.cachefunc import cached_in_parent_method, cached_method
from sage.misc.lazy_import import lazy_import
from sage.structure.unique_representation import UniqueRepresentation
from sage.structure.element_wrapper import ElementWrapper
from sage.structure.parent import Parent
from sage.categories.highest_weight_crystals import HighestWeightCrystals
from sage.categories.regular_crystals import RegularCrystals
from sage.categories.classical_crystals import ClassicalCrystals
from sage.categories.infinite_enumerated_sets import InfiniteEnumeratedSets
from sage.categories.loop_crystals import (RegularLoopCrystals,
                                           KirillovReshetikhinCrystals)
from sage.combinat.root_system.cartan_type import CartanType
from sage.rings.integer import Integer
from sage.rings.rational_field import QQ
from sage.combinat.root_system.root_system import RootSystem
from sage.arith.misc import integer_floor as floor
from sage.misc.latex import latex

lazy_import('sage.combinat.root_system.weyl_group', 'WeylGroup')


class CrystalOfLSPaths(UniqueRepresentation, Parent):
    r"""
    Crystal graph of LS paths generated from the straight-line path
    to a given weight.

    INPUT:

    - ``cartan_type`` -- (optional) the Cartan type
    - ``starting_weight`` -- a weight; if ``cartan_type`` is given,
      then the weight should be given as a list of coefficients of
      the fundamental weights, otherwise it should be given in the
      ``weight_space`` basis; for affine highest weight crystals,
      one needs to use the extended weight space

    The crystal class of piecewise linear paths in the weight space,
    generated from a straight-line path from the origin to a given
    element of the weight lattice.

    EXAMPLES::

        sage: R = RootSystem(['A',2,1])
        sage: La = R.weight_space(extended = True).basis()
        sage: B = crystals.LSPaths(La[2]-La[0]); B
        The crystal of LS paths of type ['A', 2, 1] and weight -Lambda[0] + Lambda[2]

        sage: C = crystals.LSPaths(['A',2,1],[-1,0,1]); C
        The crystal of LS paths of type ['A', 2, 1] and weight -Lambda[0] + Lambda[2]
        sage: B == C
        True
        sage: c = C.module_generators[0]; c
        (-Lambda[0] + Lambda[2],)
        sage: [c.f(i) for i in C.index_set()]
        [None, None, (Lambda[1] - Lambda[2],)]

        sage: R = C.R; R
        Root system of type ['A', 2, 1]
        sage: Lambda = R.weight_space().basis(); Lambda
        Finite family {0: Lambda[0], 1: Lambda[1], 2: Lambda[2]}
        sage: b=C(tuple([-Lambda[0]+Lambda[2]]))
        sage: b==c
        True
        sage: b.f(2)
        (Lambda[1] - Lambda[2],)

    For classical highest weight crystals, we can also compare the results
    with the tableaux implementation::

        sage: C = crystals.LSPaths(['A',2],[1,1])
        sage: sorted(C, key=str)
        [(-2*Lambda[1] + Lambda[2],), (-Lambda[1] + 1/2*Lambda[2], Lambda[1] - 1/2*Lambda[2]),
         (-Lambda[1] + 2*Lambda[2],), (-Lambda[1] - Lambda[2],),
         (1/2*Lambda[1] - Lambda[2], -1/2*Lambda[1] + Lambda[2]), (2*Lambda[1] - Lambda[2],),
         (Lambda[1] + Lambda[2],), (Lambda[1] - 2*Lambda[2],)]
        sage: C.cardinality()
        8
        sage: B = crystals.Tableaux(['A',2],shape=[2,1])
        sage: B.cardinality()
        8
        sage: B.digraph().is_isomorphic(C.digraph())
        True

    Make sure you use the weight space and not the weight lattice
    for your weights::

        sage: R = RootSystem(['A',2,1])
        sage: La = R.weight_lattice(extended = True).basis()
        sage: B = crystals.LSPaths(La[2]); B
        Traceback (most recent call last):
        ...
        ValueError: use the weight space, rather than weight lattice for your weights

    REFERENCES:

    - [Li1995b]_
    """
    @staticmethod
    def __classcall_private__(cls, starting_weight, cartan_type=None, starting_weight_parent=None):
        """
        Classcall to mend the input.

        Internally, the
        :class:`~sage.combinat.crystals.littelmann_path.CrystalOfLSPaths` code
        works with a ``starting_weight`` that is in the weight space associated
        to the crystal. The user can, however, also input a ``cartan_type``
        and the coefficients of the fundamental weights as
        ``starting_weight``. This code transforms the input into the right
        format (also necessary for UniqueRepresentation).

        TESTS::

            sage: crystals.LSPaths(['A',2,1], [-1,0,1])
            The crystal of LS paths of type ['A', 2, 1] and weight -Lambda[0] + Lambda[2]

            sage: R = RootSystem(['B',2,1])
            sage: La = R.weight_space(extended=True).basis()
            sage: C = crystals.LSPaths(['B',2,1],[0,0,1])
            sage: B = crystals.LSPaths(La[2])
            sage: B is C
            True

            sage: La = RootSystem(['A', 3]).weight_space().fundamental_weights()
            sage: crystals.LSPaths(['A', 3], La[2])
            The crystal of LS paths of type ['A', 3] and weight Lambda[2]

            sage: crystals.LSPaths(La[2] + 2*La[3], ['A', 3])
            The crystal of LS paths of type ['A', 3] and weight Lambda[2] + 2*Lambda[3]

            sage: crystals.LSPaths(La[2], starting_weight_parent=RootSystem(['B', 3]).weight_space())
            Traceback (most recent call last):
            ...
            ValueError: the passed parent is not equal to parent of the inputted weight
        """
        if cartan_type is not None:
            try:
                cartan_type, starting_weight = CartanType(starting_weight), cartan_type
            except (ValueError, TypeError):
                cartan_type = CartanType(cartan_type)
            extended = cartan_type.is_affine()

            R = RootSystem(cartan_type)
            P = R.weight_space(extended=extended)
            Lambda = P.basis()
            if not isinstance(starting_weight, P.Element):
                offset = R.index_set()[Integer(0)]
                starting_weight = P.sum(starting_weight[j-offset]*Lambda[j] for j in R.index_set())
        if starting_weight_parent is None:
            starting_weight_parent = starting_weight.parent()
        else:
            # Both the weight and the parent of the weight are passed as arguments of init to be able
            # to distinguish between crystals with the extended and non-extended weight lattice!
            if starting_weight.parent() != starting_weight_parent:
                raise ValueError("the passed parent is not equal to parent of the inputted weight")

        return super().__classcall__(cls, starting_weight, starting_weight_parent=starting_weight_parent)

    def __init__(self, starting_weight, starting_weight_parent):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: C = crystals.LSPaths(['A',2,1],[-1,0,1]); C
            The crystal of LS paths of type ['A', 2, 1] and weight -Lambda[0] + Lambda[2]
            sage: C.R
            Root system of type ['A', 2, 1]
            sage: C.weight
            -Lambda[0] + Lambda[2]
            sage: C.weight.parent()
            Extended weight space over the Rational Field of the Root system of type ['A', 2, 1]
            sage: C.module_generators
            ((-Lambda[0] + Lambda[2],),)

        TESTS::

            sage: C = crystals.LSPaths(['A',2,1], [-1,0,1])
            sage: TestSuite(C).run() # long time
            sage: C = crystals.LSPaths(['E',6], [1,0,0,0,0,0])
            sage: TestSuite(C).run()

            sage: R = RootSystem(['C',3,1])
            sage: La = R.weight_space().basis()
            sage: LaE = R.weight_space(extended=True).basis()
            sage: B = crystals.LSPaths(La[0])
            sage: BE = crystals.LSPaths(LaE[0])
            sage: B is BE
            False
            sage: B.weight_lattice_realization()
            Weight space over the Rational Field of the Root system of type ['C', 3, 1]
            sage: BE.weight_lattice_realization()
            Extended weight space over the Rational Field of the Root system of type ['C', 3, 1]
        """
        cartan_type = starting_weight.parent().cartan_type()
        self.R = RootSystem(cartan_type)
        self.weight = starting_weight
        if not self.weight.parent().base_ring().has_coerce_map_from(QQ):
            raise ValueError("use the weight space, rather than weight lattice for your weights")
        self._cartan_type = cartan_type
        if cartan_type.is_affine():
            if all(i >= 0 for i in starting_weight.coefficients()):
                Parent.__init__(self, category=(RegularCrystals(),
                                                HighestWeightCrystals(),
                                                InfiniteEnumeratedSets()))
            elif starting_weight.parent().is_extended():
                Parent.__init__(self, category=(RegularCrystals(), InfiniteEnumeratedSets()))
            else:
                cl = self._cartan_type.classical().index_set()
                if sum(self.weight[i] for i in cl) == 1:
                    cat = KirillovReshetikhinCrystals()
                else:
                    cat = RegularLoopCrystals().Finite()
                Parent.__init__(self, category=cat)
        else:
            Parent.__init__(self, category=ClassicalCrystals())

        if starting_weight == starting_weight.parent().zero():
            initial_element = self(())
        else:
            initial_element = self((starting_weight,))
        self.module_generators = (initial_element,)

    def _repr_(self):
        r"""
        EXAMPLES::

            sage: crystals.LSPaths(['B',3],[1,1,0]) # indirect doctest
            The crystal of LS paths of type ['B', 3] and weight Lambda[1] + Lambda[2]
        """
        return f"The crystal of LS paths of type {self._cartan_type} and weight {self.weight}"

    def weight_lattice_realization(self):
        r"""
        Return weight lattice realization of ``self``.

        EXAMPLES::

            sage: B = crystals.LSPaths(['B',3],[1,1,0])
            sage: B.weight_lattice_realization()
            Weight space over the Rational Field of the Root system of type ['B', 3]
            sage: B = crystals.LSPaths(['B',3,1],[1,1,1,0])
            sage: B.weight_lattice_realization()
            Extended weight space over the Rational Field of the Root system of type ['B', 3, 1]
        """
        return self.weight.parent()

    class Element(ElementWrapper):
        """
        A Littelmann path (crystal element).

        TESTS::

            sage: C = crystals.LSPaths(['E',6],[1,0,0,0,0,0])
            sage: c = C.an_element()
            sage: TestSuite(c).run()
        """
        def endpoint(self):
            r"""
            Compute the endpoint of ``self``.

            EXAMPLES::

                sage: C = crystals.LSPaths(['A',2],[1,1])
                sage: b = C.module_generators[0]
                sage: b.endpoint()
                Lambda[1] + Lambda[2]
                sage: b.f_string([1,2,2,1])
                (-Lambda[1] - Lambda[2],)
                sage: b.f_string([1,2,2,1]).endpoint()
                -Lambda[1] - Lambda[2]
                sage: b.f_string([1,2])
                (1/2*Lambda[1] - Lambda[2], -1/2*Lambda[1] + Lambda[2])
                sage: b.f_string([1,2]).endpoint()
                0
                sage: b = C([])
                sage: b.endpoint()
                0
            """
            if not self.value:
                return self.parent().weight.parent().zero()
            return sum(self.value)
            #return self.parent().R.weight_space(extended = self.parent().extended).zero()

        def compress(self):
            r"""
            Merge consecutive positively parallel steps present in ``self``.

            EXAMPLES::

                sage: C = crystals.LSPaths(['A',2],[1,1])
                sage: Lambda = C.R.weight_space().fundamental_weights(); Lambda
                Finite family {1: Lambda[1], 2: Lambda[2]}
                sage: c = C(tuple([1/2*Lambda[1]+1/2*Lambda[2], 1/2*Lambda[1]+1/2*Lambda[2]]))
                sage: c.compress()
                (Lambda[1] + Lambda[2],)
            """
            if not self.value:
                return self
            q = []
            curr = self.value[0]
            for v in self.value[1:]:
                if positively_parallel_weights(curr, v):
                    curr = curr + v
                else:
                    q.append(curr)
                    curr = v
            q.append(curr)
            return self.parent()(tuple(q))

        def split_step(self, which_step, r):
            r"""
            Split the indicated step into two parallel steps of relative
            lengths `r` and `1-r`.

            INPUT:

            - ``which_step`` -- a position in the tuple ``self``
            - ``r`` -- a rational number between 0 and 1

            EXAMPLES::

                sage: C = crystals.LSPaths(['A',2],[1,1])
                sage: b = C.module_generators[0]
                sage: b.split_step(0,1/3)
                (1/3*Lambda[1] + 1/3*Lambda[2], 2/3*Lambda[1] + 2/3*Lambda[2])
            """
            v = self.value[which_step]
            return self.parent()(self.value[:which_step] + (r*v,(1-r)*v) + self.value[which_step+1:])

        def reflect_step(self, which_step, i):
            r"""
            Apply the `i`-th simple reflection to the indicated step in ``self``.

            EXAMPLES::

                sage: C = crystals.LSPaths(['A',2],[1,1])
                sage: b = C.module_generators[0]
                sage: b.reflect_step(0,1)
                (-Lambda[1] + 2*Lambda[2],)
                sage: b.reflect_step(0,2)
                (2*Lambda[1] - Lambda[2],)
            """
            return self.parent()(self.value[:which_step]+tuple([self.value[which_step].simple_reflection(i)])+self.value[which_step+1:])

        def _string_data(self, i):
            r"""
            Compute the `i`-string data of ``self``.

            TESTS::

                sage: C = crystals.LSPaths(['A',2],[1,1])
                sage: b = C.module_generators[0]
                sage: b._string_data(1)
                ()
                sage: b._string_data(2)
                ()
                sage: b.f(1)._string_data(1)
                ((0, -1, -1),)
                sage: b.f(1).f(2)._string_data(2)
                ((0, -1, -1),)
            """
            if not self.value:
                return ()
            # get the i-th simple coroot
            alv = self.value[0].parent().alphacheck()[i]
            # Compute the i-heights of the steps of vs
            steps = [v.scalar(alv) for v in self.value]
            # Get the wet step data
            minima_pos = []
            ps = 0
            psmin = 0
            for ix, step in enumerate(steps):
                ps = ps + step
                if ps < psmin:
                    minima_pos.append((ix,ps,step))
                    psmin = ps
            return tuple(minima_pos)

        def epsilon(self, i):
            r"""
            Return the distance to the beginning of the `i`-string.

            This method overrides the generic implementation in the category of crystals
            since this computation is more efficient.

            EXAMPLES::

                sage: C = crystals.LSPaths(['A',2],[1,1])
                sage: [c.epsilon(1) for c in C]
                [0, 1, 0, 0, 1, 0, 1, 2]
                sage: [c.epsilon(2) for c in C]
                [0, 0, 1, 2, 1, 1, 0, 0]
            """
            return self.e(i, length_only=True)

        def phi(self, i):
            r"""
            Return the distance to the end of the `i`-string.

            This method overrides the generic implementation in the category of crystals
            since this computation is more efficient.

            EXAMPLES::

                sage: C = crystals.LSPaths(['A',2],[1,1])
                sage: [c.phi(1) for c in C]
                [1, 0, 0, 1, 0, 2, 1, 0]
                sage: [c.phi(2) for c in C]
                [1, 2, 1, 0, 0, 0, 0, 1]
            """
            return self.f(i, length_only=True)

        def e(self, i, power=1, to_string_end=False, length_only=False):
            r"""
            Return the `i`-th crystal raising operator on ``self``.

            INPUT:

            - ``i`` -- element of the index set of the underlying root system
            - ``power`` -- positive integer (default: 1); specifies the power
              of the raising operator to be applied
            - ``to_string_end`` -- boolean (default: ``False``); if ``True``,
              returns the dominant end of the `i`-string of ``self``
            - ``length_only`` -- boolean; if ``True``, returns the distance
              to the dominant end of the `i`-string of ``self``

            EXAMPLES::

                sage: C = crystals.LSPaths(['A',2],[1,1])
                sage: c = C[2]; c
                (1/2*Lambda[1] - Lambda[2], -1/2*Lambda[1] + Lambda[2])
                sage: c.e(1)
                sage: c.e(2)
                (-Lambda[1] + 2*Lambda[2],)
                sage: c.e(2,to_string_end=True)
                (-Lambda[1] + 2*Lambda[2],)
                sage: c.e(1,to_string_end=True)
                (1/2*Lambda[1] - Lambda[2], -1/2*Lambda[1] + Lambda[2])
                sage: c.e(1,length_only=True)
                0
            """
            data = self._string_data(i)
            # compute the minimum i-height M on the path
            if not data:
                M = 0
            else:
                M = data[-1][1]
            max_raisings = floor(-M)
            if length_only:
                return max_raisings
            # set the power of e_i to apply
            if to_string_end:
                p = max_raisings
            else:
                p = power
            if p > max_raisings:
                return None

            # copy the vector sequence into a working vector sequence ws
            #!!! ws only needs to be the actual vector sequence, not some
            #!!! fancy crystal graph element
            P = self.parent()
            ws = P(self.value)

            ix = len(data) - 1
            while ix >= 0 and data[ix][1] < M + p:
                # get the index of the current step to be processed
                j = data[ix][0]
                # find the i-height where the current step might need to be split
                if ix == 0:
                    prev_ht = M + p
                else:
                    prev_ht = min(data[ix-1][1], M+p)
                # if necessary split the step. Then reflect the wet part.
                if data[ix][1] - data[ix][2] > prev_ht:
                    ws = ws.split_step(j, 1-(prev_ht-data[ix][1])/(-data[ix][2]))
                    ws = ws.reflect_step(j+1, i)
                else:
                    ws = ws.reflect_step(j, i)
                ix -= 1
            #!!! at this point we should return the fancy crystal graph element
            #!!! corresponding to the humble vector sequence ws
            return P(ws.compress())

        def dualize(self):
            r"""
            Return the dualized path of ``self``.

            EXAMPLES::

                sage: C = crystals.LSPaths(['A',2],[1,1])
                sage: for c in C:
                ....:     print("{} {}".format(c, c.dualize()))
                (Lambda[1] + Lambda[2],) (-Lambda[1] - Lambda[2],)
                (-Lambda[1] + 2*Lambda[2],) (Lambda[1] - 2*Lambda[2],)
                (1/2*Lambda[1] - Lambda[2], -1/2*Lambda[1] + Lambda[2]) (1/2*Lambda[1] - Lambda[2], -1/2*Lambda[1] + Lambda[2])
                (Lambda[1] - 2*Lambda[2],) (-Lambda[1] + 2*Lambda[2],)
                (-Lambda[1] - Lambda[2],) (Lambda[1] + Lambda[2],)
                (2*Lambda[1] - Lambda[2],) (-2*Lambda[1] + Lambda[2],)
                (-Lambda[1] + 1/2*Lambda[2], Lambda[1] - 1/2*Lambda[2]) (-Lambda[1] + 1/2*Lambda[2], Lambda[1] - 1/2*Lambda[2])
                (-2*Lambda[1] + Lambda[2],) (2*Lambda[1] - Lambda[2],)
            """
            if not self.value:
                return self
            dual_path = [-v for v in reversed(self.value)]
            return self.parent()(tuple(dual_path))

        def f(self, i, power=1, to_string_end=False, length_only=False):
            r"""
            Return the `i`-th crystal lowering operator on ``self``.

            INPUT:

            - ``i`` -- element of the index set of the underlying root system
            - ``power`` -- positive integer (default: 1); specifies the power
              of the lowering operator to be applied
            - ``to_string_end`` -- boolean (default: ``False``); if ``True``,
              returns the anti-dominant end of the `i`-string of ``self``
            - ``length_only`` -- boolean; if ``True``, returns the distance
              to the anti-dominant end of the `i`-string of ``self``

            EXAMPLES::

                sage: C = crystals.LSPaths(['A',2],[1,1])
                sage: c = C.module_generators[0]
                sage: c.f(1)
                (-Lambda[1] + 2*Lambda[2],)
                sage: c.f(1,power=2)
                sage: c.f(2)
                (2*Lambda[1] - Lambda[2],)
                sage: c.f(2,to_string_end=True)
                (2*Lambda[1] - Lambda[2],)
                sage: c.f(2,length_only=True)
                1

                sage: C = crystals.LSPaths(['A',2,1],[-1,-1,2])
                sage: c = C.module_generators[0]
                sage: c.f(2,power=2)
                (Lambda[0] + Lambda[1] - 2*Lambda[2],)
            """
            dual_path = self.dualize()
            dual_path = dual_path.e(i, power, to_string_end, length_only)
            if length_only:
                return dual_path
            if dual_path is None:
                return None
            return dual_path.dualize()

        def s(self, i):
            r"""
            Compute the reflection of ``self`` along the `i`-string.

            This method is more efficient than the generic implementation since
            it uses powers of `e` and `f` in the Littelmann model directly.

            EXAMPLES::

                sage: C = crystals.LSPaths(['A',2],[1,1])
                sage: c = C.module_generators[0]
                sage: c.s(1)
                (-Lambda[1] + 2*Lambda[2],)
                sage: c.s(2)
                (2*Lambda[1] - Lambda[2],)

                sage: C = crystals.LSPaths(['A',2,1],[-1,0,1])
                sage: c = C.module_generators[0]; c
                (-Lambda[0] + Lambda[2],)
                sage: c.s(2)
                (Lambda[1] - Lambda[2],)
                sage: c.s(1)
                (-Lambda[0] + Lambda[2],)
                sage: c.f(2).s(1)
                (Lambda[0] - Lambda[1],)
            """
            ph = self.phi(i)
            ep = self.epsilon(i)
            diff = ph - ep
            if diff >= 0:
                return self.f(i, power=diff)
            return self.e(i, power=-diff)

        def weight(self):
            """
            Return the weight of ``self``.

            EXAMPLES::

                sage: B = crystals.LSPaths(['A',1,1],[1,0])
                sage: b = B.highest_weight_vector()
                sage: b.f(0).weight()
                -Lambda[0] + 2*Lambda[1] - delta
            """
            P = self.parent().weight_lattice_realization()
            return P.sum(self.value)

        def _latex_(self):
            r"""
            Latex method for ``self``.

            EXAMPLES::

                sage: C = crystals.LSPaths(['A',2],[1,1])
                sage: c = C.module_generators[0]
                sage: c._latex_()
                [\Lambda_{1} + \Lambda_{2}]
            """
            return [latex(p) for p in self.value]


#####################################################################
# Projected level-zero


class CrystalOfProjectedLevelZeroLSPaths(CrystalOfLSPaths):
    r"""
    Crystal of projected level zero LS paths.

    INPUT:

    - ``weight`` -- a dominant weight of the weight space of an affine
      Kac-Moody root system

    When ``weight`` is just a single fundamental weight `\Lambda_r`, this
    crystal is isomorphic to a Kirillov-Reshetikhin (KR) crystal, see also
    :meth:`sage.combinat.crystals.kirillov_reshetikhin.KirillovReshetikhinFromLSPaths`.
    For general weights, it is isomorphic to a tensor product of
    single-column KR crystals.

    EXAMPLES::

        sage: R = RootSystem(['C',3,1])
        sage: La = R.weight_space().basis()
        sage: LS = crystals.ProjectedLevelZeroLSPaths(La[1]+La[3])
        sage: LS.cardinality()
        84
        sage: GLS = LS.digraph()

        sage: K1 = crystals.KirillovReshetikhin(['C',3,1],1,1)
        sage: K3 = crystals.KirillovReshetikhin(['C',3,1],3,1)
        sage: T = crystals.TensorProduct(K3,K1)
        sage: T.cardinality()
        84
        sage: GT = T.digraph() # long time
        sage: GLS.is_isomorphic(GT, edge_labels = True) # long time
        True

    TESTS::

        sage: ct = CartanType(['A',4,2]).dual()
        sage: P = RootSystem(ct).weight_space()
        sage: La = P.fundamental_weights()
        sage: C = crystals.ProjectedLevelZeroLSPaths(La[1])
        sage: sorted(C, key=str)
        [(-Lambda[0] + Lambda[1],),
         (-Lambda[1] + 2*Lambda[2],),
         (1/2*Lambda[1] - Lambda[2], -1/2*Lambda[1] + Lambda[2]),
         (Lambda[0] - Lambda[1],),
         (Lambda[1] - 2*Lambda[2],)]
    """
    @staticmethod
    def __classcall_private__(cls, weight):
        """
        Classcall to mend the input.

        Internally, the
        :class:`~sage.combinat.crystals.littelmann_path.CrystalOfProjectedLevelZeroLSPaths`
        uses a level zero weight, which is passed on to
        :class:`~sage.combinat.crystals.littelmann_path.CrystalOfLSPaths`.
        ``weight`` is first coerced to a level zero weight.

        TESTS::

            sage: R = RootSystem(['C',3,1])
            sage: La = R.weight_space().basis()
            sage: C = crystals.ProjectedLevelZeroLSPaths(La[1] + La[2])
            sage: C2 = crystals.ProjectedLevelZeroLSPaths(La[1] + La[2])
            sage: C is C2
            True

            sage: R = RootSystem(['C',3,1])
            sage: La = R.weight_space(extended = True).basis()
            sage: crystals.ProjectedLevelZeroLSPaths(La[1] + La[2])
            Traceback (most recent call last):
            ...
            ValueError: the weight should be in the non-extended weight lattice
        """
        if weight.parent().is_extended():
            raise ValueError("the weight should be in the non-extended weight lattice")
        La = weight.parent().basis()
        weight = weight - weight.level() * La[0] / La[0].level()
        return super().__classcall__(cls, weight,
                                     starting_weight_parent=weight.parent())

    @cached_method
    def maximal_vector(self):
        """
        Return the maximal vector of ``self``.

        EXAMPLES::

            sage: R = RootSystem(['A',2,1])
            sage: La = R.weight_space().basis()
            sage: LS = crystals.ProjectedLevelZeroLSPaths(2*La[1]+La[2])
            sage: LS.maximal_vector()
            (-3*Lambda[0] + 2*Lambda[1] + Lambda[2],)
        """
        return self.module_generators[0]

    @cached_method
    def classically_highest_weight_vectors(self):
        r"""
        Return the classically highest weight vectors of ``self``.

        EXAMPLES::

            sage: R = RootSystem(['A',2,1])
            sage: La = R.weight_space().basis()
            sage: LS = crystals.ProjectedLevelZeroLSPaths(2*La[1])
            sage: LS.classically_highest_weight_vectors()
            ((-2*Lambda[0] + 2*Lambda[1],),
             (-Lambda[0] + Lambda[1], -Lambda[1] + Lambda[2]))
        """
        I0 = self.cartan_type().classical().index_set()
        return tuple([x for x in self.list() if x.is_highest_weight(I0)])

    def one_dimensional_configuration_sum(self, q=None, group_components=True):
        r"""
        Compute the one-dimensional configuration sum.

        INPUT:

        - ``q`` -- (default: ``None``) a variable or ``None``; if ``None``,
          a variable ``q`` is set in the code
        - ``group_components`` -- boolean (default: ``True``); if ``True``,
          then the terms are grouped by classical component

        The one-dimensional configuration sum is the sum of the weights
        of all elements in the crystal weighted by the energy function.
        For untwisted types it uses the parabolic quantum Bruhat graph,
        see [LNSSS2013]_. In the dual-of-untwisted case, the parabolic
        quantum Bruhat graph is defined by exchanging the roles of roots
        and coroots (which is still conjectural at this point).

        EXAMPLES::

            sage: R = RootSystem(['A',2,1])
            sage: La = R.weight_space().basis()
            sage: LS = crystals.ProjectedLevelZeroLSPaths(2*La[1])
            sage: LS.one_dimensional_configuration_sum() # long time
            B[-2*Lambda[1] + 2*Lambda[2]] + (q+1)*B[-Lambda[1]]
             + (q+1)*B[Lambda[1] - Lambda[2]] + B[2*Lambda[1]]
             + B[-2*Lambda[2]] + (q+1)*B[Lambda[2]]
            sage: R.<t> = ZZ[]
            sage: LS.one_dimensional_configuration_sum(t, False) # long time
            B[-2*Lambda[1] + 2*Lambda[2]] + (t+1)*B[-Lambda[1]]
             + (t+1)*B[Lambda[1] - Lambda[2]] + B[2*Lambda[1]]
             + B[-2*Lambda[2]] + (t+1)*B[Lambda[2]]

        TESTS::

            sage: R = RootSystem(['B',3,1])
            sage: La = R.weight_space().basis()
            sage: LS = crystals.ProjectedLevelZeroLSPaths(La[1]+La[2])
            sage: LS.one_dimensional_configuration_sum() == LS.one_dimensional_configuration_sum(group_components=False) # long time
            True
            sage: K1 = crystals.KirillovReshetikhin(['B',3,1],1,1)
            sage: K2 = crystals.KirillovReshetikhin(['B',3,1],2,1)
            sage: T = crystals.TensorProduct(K2,K1)
            sage: T.one_dimensional_configuration_sum() == LS.one_dimensional_configuration_sum() # long time
            True

            sage: R = RootSystem(['D',4,2])
            sage: La = R.weight_space().basis()
            sage: LS = crystals.ProjectedLevelZeroLSPaths(La[1]+La[2])
            sage: K1 = crystals.KirillovReshetikhin(['D',4,2],1,1)
            sage: K2 = crystals.KirillovReshetikhin(['D',4,2],2,1)
            sage: T = crystals.TensorProduct(K2,K1)
            sage: T.one_dimensional_configuration_sum() == LS.one_dimensional_configuration_sum() # long time
            True

            sage: R = RootSystem(['A',5,2])
            sage: La = R.weight_space().basis()
            sage: LS = crystals.ProjectedLevelZeroLSPaths(3*La[1])
            sage: K1 = crystals.KirillovReshetikhin(['A',5,2],1,1)
            sage: T = crystals.TensorProduct(K1,K1,K1)
            sage: T.one_dimensional_configuration_sum() == LS.one_dimensional_configuration_sum() # long time
            True
        """
        if q is None:
            from sage.rings.rational_field import QQ
            q = QQ['q'].gens()[0]
        #P0 = self.weight_lattice_realization().classical()
        P0 = RootSystem(self.cartan_type().classical()).weight_lattice()
        R = P0.base_ring()
        B = P0.algebra(q.parent())

        I0 = frozenset(P0.index_set())

        def weight(x):
            w = x.weight()
            return P0.element_class(P0, {i: R(c) for i, c in w if i in I0})

        if group_components:
            G = self.digraph(index_set=self.cartan_type().classical().index_set())
            C = G.connected_components(sort=False)
            return sum(q**(c[0].energy_function()) * B.sum(B(weight(b)) for b in c) for c in C)
        return B.sum(q**(b.energy_function()) * B(weight(b)) for b in self)

    def is_perfect(self, level=1) -> bool:
        r"""
        Check whether the crystal ``self`` is perfect (of level ``level``).

        INPUT:

        - ``level`` -- (default: 1) positive integer

        A crystal `\mathcal{B}` is perfect of level `\ell` if:

        #. `\mathcal{B}` is isomorphic to the crystal graph of a
           finite-dimensional `U_q^{'}(\mathfrak{g})`-module.
        #. `\mathcal{B}\otimes \mathcal{B}` is connected.
        #. There exists a `\lambda\in X`, such that
           `\mathrm{wt}(\mathcal{B}) \subset \lambda + \sum_{i\in I} \ZZ_{\le 0} \alpha_i`
           and there is a unique element in
           `\mathcal{B}` of classical weight `\lambda`.
        #. For all `b \in \mathcal{B}`,
           `\mathrm{level}(\varepsilon (b)) \geq \ell`.
        #. For all `\Lambda` dominant weights of level `\ell`, there exist
           unique elements `b_{\Lambda}, b^{\Lambda} \in \mathcal{B}`, such
           that `\varepsilon (b_{\Lambda}) = \Lambda = \varphi(b^{\Lambda})`.

        Points (1)-(3) are known to hold. This method checks points (4) and (5).

        EXAMPLES::

            sage: C = CartanType(['C',2,1])
            sage: R = RootSystem(C)
            sage: La = R.weight_space().basis()
            sage: LS = crystals.ProjectedLevelZeroLSPaths(La[1])
            sage: LS.is_perfect()
            False
            sage: LS = crystals.ProjectedLevelZeroLSPaths(La[2])
            sage: LS.is_perfect()
            True

            sage: C = CartanType(['E',6,1])
            sage: R = RootSystem(C)
            sage: La = R.weight_space().basis()
            sage: LS = crystals.ProjectedLevelZeroLSPaths(La[1])
            sage: LS.is_perfect()
            True
            sage: LS.is_perfect(2)
            False

            sage: C = CartanType(['D',4,1])
            sage: R = RootSystem(C)
            sage: La = R.weight_space().basis()
            sage: all(crystals.ProjectedLevelZeroLSPaths(La[i]).is_perfect() for i in [1,2,3,4])
            True

            sage: C = CartanType(['A',6,2])
            sage: R = RootSystem(C)
            sage: La = R.weight_space().basis()
            sage: LS = crystals.ProjectedLevelZeroLSPaths(La[1]+La[2])
            sage: LS.is_perfect()
            True
            sage: LS.is_perfect(2)
            False
        """
        MPhi = []
        for b in self:
            p = b.Phi().level()
            assert p == b.Epsilon().level()
            if p < level:
                return False
            if p == level:
                MPhi += [b]
        weights = []
        I = self.index_set()
        rank = len(I)
        WLR = self.weight_lattice_realization()
        R = WLR.base_ring()
        from sage.combinat.integer_vector import integer_vectors_nk_fast_iter
        for n in range(1, level+1):
            for c in integer_vectors_nk_fast_iter(n, rank):
                w = WLR.element_class(WLR, {i: R(c[i]) for i in I if c[i]})
                if w.level() == level:
                    weights.append(w)
        return sorted([b.Phi() for b in MPhi]) == sorted(weights)

    class Element(CrystalOfLSPaths.Element):
        """
        Element of a crystal of projected level zero LS paths.
        """
        @cached_in_parent_method
        def scalar_factors(self):
            r"""
            Return the scalar factors for ``self``.

            Each LS path (or ``self``) can be written as a piecewise linear map

            .. MATH::

                \pi(t) = \sum_{u'=1}^{u-1} (\sigma_{u'} - \sigma_{u'-1}) \nu_{u'}
                + (t-\sigma_{u-1}) \nu_{u}

            for `0 < \sigma_1 < \sigma_2 < \cdots < \sigma_s=1` and
            `\sigma_{u-1} \le t \le \sigma_{u}` and `1 \le u \le s`.
            This method returns the tuple of `(\sigma_1,\ldots,\sigma_s)`.

            EXAMPLES::

                sage: R = RootSystem(['C',3,1])
                sage: La = R.weight_space().basis()
                sage: LS = crystals.ProjectedLevelZeroLSPaths(La[1]+La[3])
                sage: b = LS.module_generators[0]
                sage: b.scalar_factors()
                [1]
                sage: c = b.f(1).f(3).f(2)
                sage: c.scalar_factors()
                [1/3, 1]
            """
            weight = self.parent().weight
            l = []
            s = 0
            for c in self.value:
                supp = c.support()
                if supp:
                    i = supp[0]
                    for w in weight._orbit_iter():
                        # Check whether the vectors c and w are positive scalar multiples of each other
                        # If i is not in the support of w, then the first
                        #   product is 0
                        if c[i] * w[i] > 0 and c[i] * w == w[i] * c:
                            s += c[i] / w[i]
                            l += [s]
                            break
            return l

        @cached_in_parent_method
        def weyl_group_representation(self):
            r"""
            Transform the weights in the LS path ``self`` to elements
            in the Weyl group.

            Each LS path can be written as the piecewise linear map:

            .. MATH::

                \pi(t) = \sum_{u'=1}^{u-1} (\sigma_{u'} - \sigma_{u'-1}) \nu_{u'}
                + (t-\sigma_{u-1}) \nu_{u}

            for `0 < \sigma_1 < \sigma_2 < \cdots < \sigma_s = 1` and
            `\sigma_{u-1} \le t \le \sigma_{u}` and `1 \le u \le s`.
            Each weight `\nu_u` is also associated to a Weyl group element.
            This method returns the list of Weyl group elements associated
            to the `\nu_u` for `1\le u\le s`.

            EXAMPLES::

                sage: R = RootSystem(['C',3,1])
                sage: La = R.weight_space().basis()
                sage: LS = crystals.ProjectedLevelZeroLSPaths(La[1]+La[3])
                sage: b = LS.module_generators[0]
                sage: c = b.f(1).f(3).f(2)
                sage: c.weyl_group_representation()
                [s2*s1*s3, s1*s3]
            """
            cartan = self.parent().weight.parent().cartan_type().classical()
            I = cartan.index_set()
            W = WeylGroup(cartan, prefix='s', implementation='permutation')
            return [W.from_reduced_word(x.to_dominant_chamber(index_set=I, reduced_word=True)[1]) for x in self.value]

        @cached_in_parent_method
        def energy_function(self):
            r"""
            Return the energy function of ``self``.

            The energy function `D(\pi)` of the level zero LS path
            `\pi \in \mathbb{B}_\mathrm{cl}(\lambda)` requires a series
            of definitions; for simplicity the root system is assumed to
            be untwisted affine.

            The LS path `\pi` is a piecewise linear map from the unit
            interval `[0,1]` to the weight lattice. It is specified by
            "times" `0 = \sigma_0 < \sigma_1 < \dotsm < \sigma_s = 1` and
            "direction vectors" `x_u \lambda` where `x_u \in W / W_J` for
            `1 \le u \le s`, and `W_J` is the stabilizer of `\lambda` in
            the finite Weyl group `W`. Precisely,

            .. MATH::

                \pi(t) = \sum_{u'=1}^{u-1} (\sigma_{u'}-\sigma_{u'-1})
                x_{u'} \lambda + (t-\sigma_{u-1}) x_{u} \lambda

            for `1 \le u \le s` and `\sigma_{u-1} \le t \le \sigma_{u}`.

            For any `x,y \in W / W_J`, let

            .. MATH::

                d: x = w_{0} \stackrel{\beta_{1}}{\leftarrow}
                w_{1} \stackrel{\beta_{2}}{\leftarrow} \cdots
                \stackrel{\beta_{n}}{\leftarrow} w_{n}=y

            be a shortest directed path in the parabolic quantum
            Bruhat graph. Define

            .. MATH::

                \mathrm{wt}(d) := \sum_{\substack{1 \le k \le n
                \\ \ell(w_{k-1}) < \ell(w_k)}}
                \beta_{k}^{\vee}.

            It can be shown that `\mathrm{wt}(d)` depends only on `x,y`;
            call its value `\mathrm{wt}(x,y)`. The energy function `D(\pi)`
            is defined by

            .. MATH::

                D(\pi) = -\sum_{u=1}^{s-1} (1-\sigma_{u}) \langle \lambda,
                \mathrm{wt}(x_u,x_{u+1}) \rangle.

            For more information, see [LNSSS2013]_.

            .. NOTE::

                In the dual-of-untwisted case the parabolic quantum
                Bruhat graph that is used is obtained by exchanging the
                roles of roots and coroots. Moreover, in the computation
                of the pairing the short roots must be doubled (or tripled
                for type `G`). This factor is determined by the translation
                factor of the corresponding root. Type `BC` is viewed as
                untwisted type, whereas the dual of `BC` is viewed as twisted.
                Except for the untwisted cases, these formulas are
                currently still conjectural.

            EXAMPLES::

                sage: R = RootSystem(['C',3,1])
                sage: La = R.weight_space().basis()
                sage: LS = crystals.ProjectedLevelZeroLSPaths(La[1]+La[3])
                sage: b = LS.module_generators[0]
                sage: c = b.f(1).f(3).f(2)
                sage: c.energy_function()
                0
                sage: c=b.e(0)
                sage: c.energy_function()
                1

                sage: R = RootSystem(['A',2,1])
                sage: La = R.weight_space().basis()
                sage: LS = crystals.ProjectedLevelZeroLSPaths(2*La[1])
                sage: b = LS.module_generators[0]
                sage: c = b.e(0)
                sage: c.energy_function()
                1
                sage: for c in sorted(LS, key=str):
                ....:     print("{} {}".format(c,c.energy_function()))
                (-2*Lambda[0] + 2*Lambda[1],)                    0
                (-2*Lambda[1] + 2*Lambda[2],)                    0
                (-Lambda[0] + Lambda[1], -Lambda[1] + Lambda[2]) 1
                (-Lambda[0] + Lambda[1], Lambda[0] - Lambda[2])  1
                (-Lambda[1] + Lambda[2], -Lambda[0] + Lambda[1]) 0
                (-Lambda[1] + Lambda[2], Lambda[0] - Lambda[2])  1
                (2*Lambda[0] - 2*Lambda[2],)                     0
                (Lambda[0] - Lambda[2], -Lambda[0] + Lambda[1])  0
                (Lambda[0] - Lambda[2], -Lambda[1] + Lambda[2])  0

            The next test checks that the energy function is constant
            on classically connected components::

                sage: R = RootSystem(['A',2,1])
                sage: La = R.weight_space().basis()
                sage: LS = crystals.ProjectedLevelZeroLSPaths(2*La[1]+La[2])
                sage: G = LS.digraph(index_set=[1,2])
                sage: C = G.connected_components(sort=False)
                sage: [all(c[0].energy_function()==a.energy_function() for a in c) for c in C]
                [True, True, True, True]

                sage: R = RootSystem(['D',4,2])
                sage: La = R.weight_space().basis()
                sage: LS = crystals.ProjectedLevelZeroLSPaths(La[2])
                sage: J = R.cartan_type().classical().index_set()
                sage: hw = [x for x in LS if x.is_highest_weight(J)]
                sage: [(x.weight(), x.energy_function()) for x in hw]
                [(-2*Lambda[0] + Lambda[2], 0), (-2*Lambda[0] + Lambda[1], 1), (0, 2)]
                sage: G = LS.digraph(index_set=J)
                sage: C = G.connected_components(sort=False)
                sage: [all(c[0].energy_function()==a.energy_function() for a in c) for c in C]
                [True, True, True]

                sage: R = RootSystem(CartanType(['G',2,1]).dual())
                sage: La = R.weight_space().basis()
                sage: LS = crystals.ProjectedLevelZeroLSPaths(La[1]+La[2])
                sage: G = LS.digraph(index_set=[1,2])
                sage: C = G.connected_components(sort=False)
                sage: [all(c[0].energy_function()==a.energy_function() for a in c) for c in C] # long time
                [True, True, True, True, True, True, True, True, True, True, True, True, True, True, True, True]

                sage: ct = CartanType(['BC',2,2]).dual()
                sage: R = RootSystem(ct)
                sage: La = R.weight_space().basis()
                sage: LS = crystals.ProjectedLevelZeroLSPaths(2*La[1]+La[2])
                sage: G = LS.digraph(index_set=R.cartan_type().classical().index_set())
                sage: C = G.connected_components(sort=False)
                sage: [all(c[0].energy_function()==a.energy_function() for a in c) for c in C] # long time
                [True, True, True, True, True, True, True, True, True, True, True]

                sage: R = RootSystem(['BC',2,2])
                sage: La = R.weight_space().basis()
                sage: LS = crystals.ProjectedLevelZeroLSPaths(2*La[1]+La[2])
                sage: G = LS.digraph(index_set=R.cartan_type().classical().index_set())
                sage: C = G.connected_components(sort=False)
                sage: [all(c[0].energy_function()==a.energy_function() for a in c) for c in C] # long time
                [True, True, True, True, True, True, True, True, True, True, True, True, True, True, True,
                True, True, True, True, True, True, True, True, True, True, True, True, True, True, True, True]
            """
            weight = self.parent().weight
            P = weight.parent()
            c_weight = P.classical()(weight)
            ct = P.cartan_type()
            cartan = ct.classical()
            Qv = RootSystem(cartan).coroot_lattice()
            W = WeylGroup(cartan, prefix='s', implementation='permutation')
            J = tuple(weight.weyl_stabilizer())
            L = self.weyl_group_representation()
            if ct.is_untwisted_affine() or ct.type() == 'BC':
                untwisted = True
                G = W.quantum_bruhat_graph(J)
            else:
                untwisted = False
                cartan_dual = cartan.dual()
                Wd = WeylGroup(cartan_dual, prefix='s', implementation='permutation')
                G = Wd.quantum_bruhat_graph(J)
                Qd = RootSystem(cartan_dual).root_lattice()

                def dualize(x):
                    return Qv.from_vector(x.to_vector())

                L = [Wd.from_reduced_word(x.reduced_word()) for x in L]

                def stretch_short_root(a):
                    # stretches roots by translation factor
                    if ct.dual().type() == 'BC':
                        return ct.c()[a.to_simple_root()]*a
                    return ct.dual().c()[a.to_simple_root()]*a
                    #if a.is_short_root():
                    #    if cartan_dual.type() == 'G':
                    #        return 3*a
                    #    else:
                    #        return 2*a
                    #return a

            paths = [G.shortest_path(L[i+1],L[i]) for i in range(len(L)-1)]
            paths_labels = [[G.edge_label(p[i],p[i+1]) for i in range(len(p)-1) if p[i].length()+1 != p[i+1].length()] for p in paths]
            scalars = self.scalar_factors()
            if untwisted:
                s = sum((1 - scalars[i]) * c_weight.scalar(Qv.sum(root.associated_coroot() for root in label))
                        for i, label in enumerate(paths_labels))
                if ct.type() == 'BC':
                    return 2 * s
                return s
            s = sum((1 - scalars[i]) * c_weight.scalar(dualize(Qd.sum(stretch_short_root(root) for root in label)))
                    for i, label in enumerate(paths_labels))
            if ct.dual().type() == 'BC':
                return s / 2
            return s


#####################################################################
# B(\infty)


class InfinityCrystalOfLSPaths(UniqueRepresentation, Parent):
    r"""
    LS path model for `\mathcal{B}(\infty)`.

    Elements of `\mathcal{B}(\infty)` are equivalence classes of paths `[\pi]`
    in `\mathcal{B}(k\rho)` for `k\gg 0`, where `\rho` is the Weyl vector.  A
    canonical representative for an element of `\mathcal{B}(\infty)` is chosen
    by taking `k` to be minimal such that the endpoint of `\pi` is strictly
    dominant but its representative in `\mathcal{B}((k-1)\rho)` is on the wall
    of the dominant chamber.

    REFERENCES:

    - [LZ2011]_
    """
    @staticmethod
    def __classcall_private__(cls, cartan_type):
        """
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: B1 = crystals.infinity.LSPaths(['A',4])
            sage: B2 = crystals.infinity.LSPaths('A4')
            sage: B3 = crystals.infinity.LSPaths(CartanType(['A',4]))
            sage: B1 is B2 and B2 is B3
            True
        """
        cartan_type = CartanType(cartan_type)
        return super().__classcall__(cls, cartan_type)

    def __init__(self, cartan_type):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: B = crystals.infinity.LSPaths(['D',4,3])
            sage: TestSuite(B).run(max_runs=500)
            sage: B = crystals.infinity.LSPaths(['B',3])
            sage: TestSuite(B).run() # long time
        """
        Parent.__init__(self, category=(HighestWeightCrystals(),
                                        InfiniteEnumeratedSets()))
        self._cartan_type = cartan_type
        self.module_generators = (self.module_generator(),)

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: crystals.infinity.LSPaths(['A',4])
            The infinity crystal of LS paths of type ['A', 4]
        """
        return "The infinity crystal of LS paths of type %s" % self._cartan_type

    @cached_method
    def module_generator(self):
        r"""
        Return the module generator (or highest weight element) of ``self``.

        The module generator is the unique path
        `\pi_\infty\colon t \mapsto t\rho`, for `t \in [0,\infty)`.

        EXAMPLES::

            sage: B = crystals.infinity.LSPaths(['A',6,2])
            sage: mg = B.module_generator(); mg
            (Lambda[0] + Lambda[1] + Lambda[2] + Lambda[3],)
            sage: mg.weight()
            0
        """
        rho = self.weight_lattice_realization().rho()
        return self((rho,))

    def weight_lattice_realization(self):
        """
        Return the weight lattice realization of ``self``.

        EXAMPLES::

            sage: B = crystals.infinity.LSPaths(['C',4])
            sage: B.weight_lattice_realization()
            Weight space over the Rational Field of the Root system of type ['C', 4]
        """
        if self._cartan_type.is_affine():
            return self._cartan_type.root_system().weight_space(extended=True)
        return self._cartan_type.root_system().weight_space()

    class Element(CrystalOfLSPaths.Element):
        def e(self, i, power=1, length_only=False):
            r"""
            Return the `i`-th crystal raising operator on ``self``.

            INPUT:

            - ``i`` -- element of the index set
            - ``power`` -- (default: 1) positive integer; specifies the
              power of the lowering operator to be applied
            - ``length_only`` -- boolean (default: ``False``); if ``True``,
              then return the distance to the anti-dominant end of the
              `i`-string of ``self``

            EXAMPLES::

                sage: B = crystals.infinity.LSPaths(['B',3,1])
                sage: mg = B.module_generator()
                sage: mg.e(0)
                sage: mg.e(1)
                sage: mg.e(2)
                sage: x = mg.f_string([1,0,2,1,0,2,1,1,0])
                sage: all(x.f(i).e(i) == x for i in B.index_set())
                True
                sage: all(x.e(i).f(i) == x for i in B.index_set() if x.epsilon(i) > 0)
                True

            TESTS:

            Check that this works in affine types::

                sage: B = crystals.infinity.LSPaths(['A',3,1])
                sage: mg = B.highest_weight_vector()
                sage: x = mg.f_string([0,1,2,3])
                sage: x.e_string([3,2,1,0]) == mg
                True

            We check that :meth:`epsilon` works::

                sage: B = crystals.infinity.LSPaths(['D',4])
                sage: mg = B.highest_weight_vector()
                sage: x = mg.f_string([1,3,4,2,4,3,2,1,4])
                sage: [x.epsilon(i) for i in B.index_set()]
                [1, 1, 0, 1]

            Check that :issue:`21671` is fixed::

                sage: B = crystals.infinity.LSPaths(['G',2])
                sage: len(B.subcrystal(max_depth=7))
                116
            """
            ret = super().e(i, power=power,
                            length_only=length_only)
            if ret is None:
                return None
            if length_only:
                return ret
            WLR = self.parent().weight_lattice_realization()
            value = list(ret.value)
            endpoint = sum(value)
            rho = WLR.rho()
            h = WLR.simple_coroots()

            if not positively_parallel_weights(value[-1], rho):
                value.append(rho)
                endpoint += rho

            while any(endpoint.scalar(alc) < 1 for alc in h):
                value[-1] += rho
                endpoint += rho
            while all(endpoint.scalar(alc) > 1 for alc in h) and value[-1] != WLR.zero():
                value[-1] -= rho
                endpoint -= rho
            while value[-1] == WLR.zero():
                value.pop()
            ret.value = tuple(value)
            return ret

        def f(self, i, power=1, length_only=False):
            r"""
            Return the `i`-th crystal lowering operator on ``self``.

            INPUT:

            - ``i`` -- element of the index set
            - ``power`` -- (default: 1) positive integer; specifies the
              power of the lowering operator to be applied
            - ``length_only`` -- boolean (default: ``False``); if ``True``,
              then return the distance to the anti-dominant end of the
              `i`-string of ``self``

            EXAMPLES::

                sage: B = crystals.infinity.LSPaths(['D',3,2])
                sage: mg = B.highest_weight_vector()
                sage: mg.f(1)
                (3*Lambda[0] - Lambda[1] + 3*Lambda[2],
                 2*Lambda[0] + 2*Lambda[1] + 2*Lambda[2])
                sage: mg.f(2)
                (Lambda[0] + 2*Lambda[1] - Lambda[2],
                 2*Lambda[0] + 2*Lambda[1] + 2*Lambda[2])
                sage: mg.f(0)
                (-Lambda[0] + 2*Lambda[1] + Lambda[2] - delta,
                 2*Lambda[0] + 2*Lambda[1] + 2*Lambda[2])
            """
            dual_path = self.dualize()
            dual_path = super(InfinityCrystalOfLSPaths.Element, dual_path).e(i, power, length_only=length_only)
            if length_only:
                return dual_path
            if dual_path is None:
                return None
            ret = dual_path.dualize()
            WLR = self.parent().weight_lattice_realization()
            value = list(ret.value)
            endpoint = sum(value)
            rho = WLR.rho()
            h = WLR.simple_coroots()

            if not positively_parallel_weights(value[-1], rho):
                value.append(rho)
                endpoint += rho

            while any(endpoint.scalar(alc) < 1 for alc in h):
                value[-1] += rho
                endpoint += rho
            while all(endpoint.scalar(alc) > 1 for alc in h) and value[-1] != WLR.zero():
                value[-1] -= rho
                endpoint -= rho
            while value[-1] == WLR.zero():
                value.pop()
            ret.value = tuple(value)
            return ret

        @cached_method
        def weight(self):
            """
            Return the weight of ``self``.

            .. TODO::

                This is a generic algorithm. We should find a better
                description and implement it.

            EXAMPLES::

                sage: B = crystals.infinity.LSPaths(['E',6])
                sage: mg = B.highest_weight_vector()
                sage: f_seq = [1,4,2,6,4,2,3,1,5,5]
                sage: x = mg.f_string(f_seq)
                sage: x.weight()
                -3*Lambda[1] - 2*Lambda[2] + 2*Lambda[3] + Lambda[4] - Lambda[5]

                sage: al = B.cartan_type().root_system().weight_space().simple_roots()
                sage: x.weight() == -sum(al[i] for i in f_seq)
                True
            """
            WLR = self.parent().weight_lattice_realization()
            alpha = WLR.simple_roots()
            return -WLR.sum(alpha[i] for i in self.to_highest_weight()[1])

        def phi(self, i):
            r"""
            Return `\varphi_i` of ``self``.

            Let `\pi \in \mathcal{B}(\infty)`. Define

            .. MATH::

                \varphi_i(\pi) := \varepsilon_i(\pi) + \langle h_i,
                \mathrm{wt}(\pi) \rangle,

            where `h_i` is the `i`-th simple coroot and `\mathrm{wt}(\pi)`
            is the :meth:`weight` of `\pi`.

            INPUT:

            - ``i`` -- element of the index set

            EXAMPLES::

                sage: B = crystals.infinity.LSPaths(['D',4])
                sage: mg = B.highest_weight_vector()
                sage: x = mg.f_string([1,3,4,2,4,3,2,1,4])
                sage: [x.phi(i) for i in B.index_set()]
                [-1, 4, -2, -3]
            """
            WLR = self.parent().weight_lattice_realization()
            h = WLR.simple_coroots()
            return self.epsilon(i) + WLR(self.weight()).scalar(h[i])


#####################################################################
# Helper functions


def positively_parallel_weights(v, w):
    """
    Check whether the vectors ``v`` and ``w`` are positive scalar
    multiples of each other.

    EXAMPLES::

        sage: from sage.combinat.crystals.littelmann_path import positively_parallel_weights
        sage: La = RootSystem(['A',5,2]).weight_space(extended=True).fundamental_weights()
        sage: rho = sum(La)
        sage: positively_parallel_weights(rho, 4*rho)
        True
        sage: positively_parallel_weights(4*rho, rho)
        True
        sage: positively_parallel_weights(rho, -rho)
        False
        sage: positively_parallel_weights(rho, La[1] + La[2])
        False
    """
    supp = v.support()
    if supp:
        i = supp[0]
        if v[i] * w[i] > 0 and v[i] * w == w[i] * v:
            return True
    return False



r"""
Interface to Scilab

Scilab is a scientific software package for numerical computations
providing a powerful open computing environment for engineering and
scientific applications.  Scilab includes hundreds of mathematical
functions with the possibility to add interactively programs from
various languages (C, C++, Fortran...).  It has sophisticated data
structures (including lists, polynomials, rational functions, linear
systems...), an interpreter and a high level programming language.

The commands in this section only work if you have the "scilab"
interpreter installed and available in your PATH.  It's not necessary
to install any special Sage packages.

EXAMPLES::

    sage: # optional - scilab
    sage: scilab.eval('2+2')
    'ans  =\n \n    4.'
    sage: scilab('2+2')
    4.
    sage: a = scilab(10)
    sage: a**10
    1.000D+10

Tutorial based the MATLAB interface tutorial:

EXAMPLES::

    sage: # optional - scilab
    sage: scilab('4+10')
    14.
    sage: scilab('date')
    15-Feb-2010
    sage: scilab('5*10 + 6')
    56.
    sage: scilab('(6+6)/3')
    4.
    sage: scilab('9')^2
    81.
    sage: a = scilab(10); b = scilab(20); c = scilab(30)
    sage: avg = (a+b+c)/3
    sage: avg
    20.
    sage: parent(avg)
    Scilab

    sage: # optional - scilab
    sage: my_scalar = scilab('3.1415')
    sage: my_scalar
    3.1415
    sage: my_vector1 = scilab('[1,5,7]')
    sage: my_vector1
    1.    5.    7.
    sage: my_vector2 = scilab('[1;5;7]')
    sage: my_vector2
    1.
    5.
    7.
    sage: my_vector1 * my_vector2
    75.

    sage: # optional - scilab
    sage: row_vector1 = scilab('[1 2 3]')
    sage: row_vector2 = scilab('[3 2 1]')
    sage: matrix_from_row_vec = scilab('[%s; %s]'%(row_vector1.name(), row_vector2.name()))
    sage: matrix_from_row_vec
    1.    2.    3.
    3.    2.    1.

    sage: # optional - scilab
    sage: column_vector1 = scilab('[1;3]')
    sage: column_vector2 = scilab('[2;8]')
    sage: matrix_from_col_vec = scilab('[%s %s]'%(column_vector1.name(), column_vector2.name()))
    sage: matrix_from_col_vec
    1.    2.
    3.    8.

    sage: my_matrix = scilab('[8, 12, 19; 7, 3, 2; 12, 4, 23; 8, 1, 1]')    # optional - scilab
    sage: my_matrix                                      # optional - scilab
        8.     12.    19.
        7.     3.     2.
        12.    4.     23.
        8.     1.     1.

    sage: combined_matrix = scilab('[%s, %s]'%(my_matrix.name(), my_matrix.name()))                                        # optional - scilab
    sage: combined_matrix                               # optional - scilab
    8.     12.    19.    8.     12.    19.
    7.     3.     2.     7.     3.     2.
    12.    4.     23.    12.    4.     23.
    8.     1.     1.     8.     1.     1.

    sage: tm = scilab('0.5:2:10')                       # optional - scilab
    sage: tm                                            # optional - scilab
    0.5    2.5    4.5    6.5    8.5

    sage: # optional - scilab
    sage: my_vector1 = scilab('[1,5,7]')
    sage: my_vector1(1)
    1.
    sage: my_vector1(2)
    5.
    sage: my_vector1(3)
    7.

Matrix indexing works as follows::

    sage: my_matrix = scilab('[8, 12, 19; 7, 3, 2; 12, 4, 23; 8, 1, 1]')     # optional - scilab
    sage: my_matrix(3,2)                                # optional - scilab
    4.

One can also use square brackets::

    sage: my_matrix[3,2]                                # optional - scilab
    4.


Setting using parenthesis cannot work (because of how the Python
language works).  Use square brackets or the set function::

    sage: # optional - scilab
    sage: my_matrix = scilab('[8, 12, 19; 7, 3, 2; 12, 4, 23; 8, 1, 1]')
    sage: my_matrix.set(2,3, 1999)
    sage: my_matrix
           8.         12.         19.
           7.          3.       1999.
          12.          4.         23.
           8.          1.          1.
    sage: my_matrix[2,3] = -126
    sage: my_matrix
           8.         12.         19.
           7.          3.      - 126.
          12.          4.         23.
           8.          1.          1.

TESTS::

    sage: # optional - scilab
    sage: M = scilab(x)
    Traceback (most recent call last):
    ...
    TypeError: ..._interface_init_() takes exactly one argument (0 given)
    sage: M = scilab(matrix(3,range(9))); M
        0.    1.    2.
        3.    4.    5.
        6.    7.    8.
    sage: M(10)
    Traceback (most recent call last):
    ...
    TypeError: Error executing code in Scilab
    ...
    Invalid index.
    sage: M[10]
    Traceback (most recent call last):
    ...
    TypeError: Error executing code in Scilab
    ...
    Invalid index.
    sage: M(4,2)
    Traceback (most recent call last):
    ...
    TypeError: Error executing code in Scilab
    ...
    Invalid index.
    sage: M[2,4]
    Traceback (most recent call last):
    ...
    TypeError: Error executing code in Scilab
    ...
    Invalid index.
    sage: M(9) = x
    Traceback (most recent call last):
    ...
    SyntaxError: can...t assign to function call (..., line 1)

AUTHORS:

- Ronan Paixao (2008-11-26), based on the MATLAB tutorial by
  William Stein (2006-10-11)
"""
# ****************************************************************************
#       Copyright (C) 2006 William Stein <wstein@gmail.com>
#       Copyright (C) 2008 Ronan Paixao <ronanpaixao@yahoo.com.br>
#
#  Distributed under the terms of the GNU General Public License (GPL).
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import os

from sage.interfaces.expect import Expect, ExpectElement
from sage.misc.instancedoc import instancedoc


class Scilab(Expect):
    """
    Interface to the Scilab interpreter.

    EXAMPLES::

        sage: # optional - scilab
        sage: a = scilab('[ 1, 1, 2; 3, 5, 8; 13, 21, 33 ]')
        sage: b = scilab('[ 1; 3; 13]')
        sage: c = a * b
        sage: print(c)
          30.
          122.
          505.
    """
    def __init__(self, maxread=None, script_subdirectory=None,
                 logfile=None, server=None, server_tmpdir=None,
                 seed=None):
        """
        Initialize the Scilab class.

        EXAMPLES::

            sage: from sage.interfaces.scilab import Scilab
            sage: sci_obj = Scilab()
            sage: del sci_obj
        """
        Expect.__init__(self,
                        name='scilab',
                        prompt='-->',
                        command="scilab -nw",
                        server=server,
                        server_tmpdir=server_tmpdir,
                        script_subdirectory=script_subdirectory,
                        restart_on_ctrlc=False,
                        verbose_start=False,
                        logfile=logfile,
                        eval_using_file_cutoff=100)
        self._seed = seed

    def set_seed(self, seed=None):
        """
        Set the seed for gp interpreter.

        The seed should be an integer.

        EXAMPLES::

            sage: # optional - scilab
            sage: from sage.interfaces.scilab import Scilab
            sage: s = Scilab()
            sage: s.set_seed(1)
            1
            sage: [s.rand() for i in range(5)]
            [
            <BLANKLINE>
                 0.6040239,
            <BLANKLINE>
                 0.0079647,
            <BLANKLINE>
                 0.6643966,
            <BLANKLINE>
                 0.9832111,
            <BLANKLINE>
                 0.5321420]
        """
        if seed is None:
            seed = self.rand_seed()
        self.eval("rand('seed',%d)" % seed)
        self._seed = seed
        return seed

    def _quit_string(self):
        """
        Return the string used to quit the pexpect interface.

        EXAMPLES::

            sage: scilab._quit_string()                 # optional - scilab
            'quit;'
        """
        return 'quit;'

    def _install_hints(self):
        """
        Hints for installing Scilab.

        EXAMPLES::

            sage: print(scilab._install_hints())       # optional - scilab
            You must ...
        """
        return """
        You must obtain the Scilab program in order to use Scilab
        from Sage.   You can read all about Scilab at
                  http://www.scilab.org/
        The executable must be accessible system-wide.
        """

    def _start(self):
        """
        Starts Scilab and sets some options.

        EXAMPLES::

            sage: scilab._start()                       # optional - scilab
        """
        Expect._start(self)
        self.eval("mode(0)")

        # set random seed
        self.set_seed(self._seed)

    def eval(self, command, *args, **kwds):
        """
        Evaluates commands.

        EXAMPLES::

            sage: scilab.eval("5")                      # optional - scilab
            'ans  =\n \n    5.'
            sage: scilab.eval("d=44")                   # optional - scilab
            'd  =\n \n    44.'
        """
        s = Expect.eval(self, command, **kwds)
        return s.replace("\x1b[?1l\x1b>", "").strip()

    def whos(self, name=None, typ=None):
        """
        Return information about current objects.
        Arguments:
        nam: first characters of selected names
        typ: name of selected Scilab variable type

        EXAMPLES::

            sage: scilab.whos("core")                   # optional - scilab
            'Name                     Type           Size           Bytes...'
            sage: scilab.whos(typ='function')           # optional - scilab
            'Name                     Type           Size           Bytes...'
        """
        parameters = ""
        if name:
            parameters += " -name %s" % (str(name))
        if typ:
            parameters += " -type %s" % (str(typ))
        return self.eval('whos' + parameters)

    def set(self, var, value):
        """
        Set the variable var to the given value.

        EXAMPLES::

            sage: scilab.set('a', 123)        # optional - scilab
            sage: scilab.get('a')               # optional - scilab
            '\n \n    123.'
        """
        cmd = '%s=%s;' % (var, value)
        out = self.eval(cmd)
        if out.find("error") != -1:
            raise TypeError("Error executing code in Scilab\nCODE:\n\t%s\nScilab ERROR:\n\t%s" % (cmd, out))

    def get(self, var):
        """
        Get the value of the variable ``var``.

        EXAMPLES::

            sage: scilab.eval('b=124;')                 # optional - scilab
            ''
            sage: scilab.get('b')                       # optional - scilab
            '\n \n    124.'
        """
        s = self.eval(f'{var}')
        i = s.find('=')
        return s[i+1:]

    def console(self):
        """
        Starts Scilab console.

        EXAMPLES::

            sage: scilab.console()          # optional - scilab; not tested
        """
        scilab_console()

    def version(self):
        """
        Return the version of the Scilab software used.

        EXAMPLES::

            sage: scilab.version()                      # optional - scilab
            'scilab-...'
        """
        return scilab_version()

    def sage2scilab_matrix_string(self, A):
        """
        Return a Scilab matrix from a Sage matrix.

        INPUT:

        - ``A`` -- Sage matrix with entries in the rationals or reals

        OUTPUT: string that evaluates to a Scilab matrix

        EXAMPLES::

            sage: M33 = MatrixSpace(QQ,3,3)             # optional - scilab
            sage: A = M33([1,2,3,4,5,6,7,8,0])          # optional - scilab
            sage: scilab.sage2scilab_matrix_string(A)   # optional - scilab
            '[1, 2, 3; 4, 5, 6; 7, 8, 0]'
        """
        s = str(A.rows())
        return s.replace('), (', '; ').replace('(', '').replace(')', '')

    def _object_class(self):
        """
        Return the class of the object.

        EXAMPLES::

            sage: scilab._object_class()                # optional - scilab
            <class 'sage.interfaces.scilab.ScilabElement'>
            sage: type(scilab(2))                       # optional - scilab
            <class 'sage.interfaces.scilab.ScilabElement'>
        """
        return ScilabElement


@instancedoc
class ScilabElement(ExpectElement):
    def __getitem__(self, n):
        """
        Use parenthesis for Scilab matrices instead.

        EXAMPLES::

            sage: # optional - scilab
            sage: M = scilab('[1,2,3;4,5,6;7,8,9]')
            sage: M[1]
            1.
            sage: M[7]
            3.
            sage: M[3,2]
            8.
        """
        if isinstance(n, tuple):
            index = str(n)[1:-1]
        else:
            index = str(n)
        return self.parent()('%s(%s)' % (self._name, index))

    def __setitem__(self, n, value):
        """
        Set an element of a matrix.

        EXAMPLES::

            sage: # optional - scilab
            sage: M = scilab('[1,2,3;4,5,6;7,8,9]')
            sage: M[6] = 0
            sage: M
            1.    2.    3.
            4.    5.    6.
            7.    0.    9.
            sage: M[3,2] = 10
            sage: M
            1.    2.    3.
            4.    5.    6.
            7.    10.    9.
        """
        if isinstance(n, tuple):
            index = str(n)[1:-1]
        else:
            index = str(n)
        self.parent().eval('%s(%s) = %s' % (self._name, index, value))

    def _matrix_(self, R):
        r"""
        Return \sage matrix from this scilab element.

        EXAMPLES::

            sage: # optional - scilab
            sage: A = scilab('[1,2;3,4]')
            sage: matrix(ZZ, A)
            [1 2]
            [3 4]
            sage: A = scilab('[1,2;3,4.5]')
            sage: matrix(RR, A)
            [1.00000000000000 2.00000000000000]
            [3.00000000000000 4.50000000000000]
        """
        from sage.matrix.matrix_space import MatrixSpace
        s = str(self).strip()
        v = s.split('\n ')
        nrows = len(v)
        if nrows == 0:
            return MatrixSpace(R, 0, 0)(0)
        ncols = len(v[0].split())
        M = MatrixSpace(R, nrows, ncols)
        v = sum([[x.rstrip('.') for x in w.split()] for w in v], [])
        return M(v)

    def set(self, i, j, x):
        """
        Set the variable var to the given value.

        EXAMPLES::

            sage: scilab.set('c', 125)          # optional - scilab
            sage: scilab.get('c')               # optional - scilab
            '\n \n    125.'
        """
        P = self._check_valid()
        z = P(x)
        P.eval('%s(%s,%s) = %s' % (self.name(), i, j, z.name()))


# An instance
scilab = Scilab()


def scilab_console():
    """
    This requires that the optional Scilab program be installed and in
    your PATH, but no optional Sage packages need to be installed.

    EXAMPLES::

        sage: from sage.interfaces.scilab import scilab_console  # optional - scilab
        sage: scilab_console()                               # optional - scilab; not tested
                ___________________________________________
                               scilab-5.0.3

                         Consortium Scilab (DIGITEO)
                       Copyright (c) 1989-2008 (INRIA)
                       Copyright (c) 1989-2007 (ENPC)
                ___________________________________________


        Startup execution:
          loading initial environment

        -->2+3
        ans  =

           5.

        -->quit

    Typing quit exits the Scilab console and returns you to Sage.
    Scilab, like Sage, remembers its history from one session to
    another.
    """
    os.system('scilab -nw')


def scilab_version():
    """
    Return the version of Scilab installed.

    EXAMPLES::

        sage: from sage.interfaces.scilab import scilab_version  # optional - scilab
        sage: scilab_version()    # optional - scilab
        'scilab-...'
    """
    return str(scilab('getversion()')).strip()



r"""
Morphic words

This modules implements morphic words (letter-to-letter coding of fixed
point of a morphism).

AUTHORS:

- Jana Lepsova (January 2021): initial version

EXAMPLES:

Creation of the fixed point of a morphism::

    sage: m = WordMorphism('a->abc,b->baba,c->ca')
    sage: w = m.fixed_point('a'); w
    word: abcbabacababaabcbabaabccaabcbabaabcbabaa...
    sage: w.length()
    +Infinity

Computing the `n`-th letter of a fixed point is fast as it is using the
abstract numeration system associated to the morphism and the starting
letter, see chapter 3 of the book [BR2010b]_::

    sage: w[10000000]                                                                   # needs sage.modules
    'b'
"""
from collections.abc import Iterator
from itertools import chain

from sage.combinat.words.word_infinite_datatypes import WordDatatype_callable
from sage.misc.lazy_import import lazy_import
from sage.rings.infinity import Infinity

lazy_import('sage.modules.free_module_element', 'vector')


class WordDatatype_morphic(WordDatatype_callable):
    r"""
    Datatype for a morphic word defined by a morphism, a starting letter
    and a coding.
    """
    def __init__(self, parent, morphism, letter,
                 coding=None, length=Infinity) -> None:
        r"""
        INPUT:

        - ``parent`` -- a parent
        - ``morphism`` -- a word morphism
        - ``letter`` -- a starting letter
        - ``coding`` -- dictionary (default: ``None``); if ``None``
          the identity map is used for the coding
        - ``length`` -- integer or ``'finite'`` or ``Infinity`` or
          ``'unknown'`` (default: ``Infinity``) the length of the word

        EXAMPLES::

            sage: m = WordMorphism('a->ab,b->a')
            sage: w = m.fixed_point('a')
            sage: w
            word: abaababaabaababaababaabaababaabaababaaba...
            sage: w[555:1000]                                                           # needs sage.modules
            word: abaababaabaababaababaabaababaabaababaaba...
            sage: w.length()
            +Infinity

        ::

            sage: m = WordMorphism('a->abc,b->baba,c->ca')
            sage: m.fixed_point('a')
            word: abcbabacababaabcbabaabccaabcbabaabcbabaa...
            sage: w = m.fixed_point('a')
            sage: w[7]                                                                  # needs sage.modules
            'c'
            sage: w[2:7]                                                                # needs sage.modules
            word: cbaba
            sage: w[500:503]                                                            # needs sage.modules
            word: caa

        When the morphic word is finite::

            sage: m = WordMorphism("a->ab,b->")
            sage: w = m.fixed_point("a"); w
            word: ab
            sage: w[0]                                                                  # needs sage.modules
            'a'
            sage: w.length()
            2

        Using the coding argument::

            sage: m = WordMorphism('a->ab,b->a')
            sage: W = m.domain()
            sage: from sage.combinat.words.morphic import WordDatatype_morphic
            sage: coding = {'a':'x', 'b':'y'}
            sage: w = WordDatatype_morphic(W, m, 'a', coding=coding)
            sage: [w[i] for i in range(10)]                                             # needs sage.modules
            ['x', 'y', 'x', 'x', 'y', 'x', 'y', 'x', 'x', 'y']

        TESTS::

            sage: m = WordMorphism('a->abcd,b->bbc,c->cddd,d->cba')
            sage: w = m.fixed_point('a')
            sage: it = iter(w)
            sage: for _ in range(10000): _ = next(it)
            sage: L = [next(it) for _ in range(10)]; L
            ['d', 'd', 'd', 'c', 'd', 'd', 'd', 'c', 'b', 'a']
            sage: w[10000:10010]                                                        # needs sage.modules
            word: dddcdddcba
            sage: list(w[10000:10010]) == L                                             # needs sage.modules
            True
        """
        self._parent = parent
        # self._func = callable
        # for hashing
        self._hash = None

        if length is Infinity:
            self._len = Infinity
        elif length is None or length == 'unknown' or length == 'finite':
            self._len = None
        else:
            self._len = length

        self._morphism = morphism
        self._letter = letter
        self._alphabet = self._morphism.domain().alphabet()
        if coding is None:
            self._coding = {a: a for a in self._alphabet}
        else:
            self._coding = coding

    def __reduce__(self) -> tuple:
        r"""
        EXAMPLES::

            sage: m = WordMorphism('a->ab,b->a')
            sage: w = m.fixed_point('a')
            sage: w.__reduce__()
            (<class 'sage.combinat.words.word.InfiniteWord_morphic'>,
             (Infinite words over {'a', 'b'},
              WordMorphism: a->ab, b->a,
              'a',
              {'a': 'a', 'b': 'b'},
              +Infinity))

        Below is the behavior for words of finite length::

            sage: m = WordMorphism("a->ab,b->")
            sage: w = m.fixed_point("a")
            sage: w.__reduce__()
            (<class 'sage.combinat.words.word.FiniteWord_morphic'>,
             (Finite words over {'a', 'b'},
              WordMorphism: a->ab, b->,
              'a',
              {'a': 'a', 'b': 'b'},
              2))
        """
        return self.__class__, (self._parent, self._morphism, self._letter,
                                self._coding, self._len)

    def representation(self, n) -> list:
        r"""
        Return the representation of the integer n in the numeration system
        associated to the morphism.

        INPUT:

        - ``n`` -- nonnegative integer

        OUTPUT: list

        EXAMPLES::

            sage: m = WordMorphism('a->ab,b->a')
            sage: w = m.fixed_point('a')
            sage: w.representation(5)                                                   # needs sage.modules
            [1, 0, 0, 0]

        When the morphic word is finite::

            sage: m = WordMorphism("a->ab,b->,c->cdab,d->dcab")
            sage: w = m.fixed_point("a")
            sage: w.representation(0)                                                   # needs sage.modules
            []
            sage: w.representation(1)                                                   # needs sage.modules
            [1]
            sage: w.representation(2)                                                   # needs sage.modules
            Traceback (most recent call last):
            ...
            IndexError: index (=2) out of range, the fixed point is finite and has length 2

        TESTS:

        Accessing this method from an instance of the current class (no using
        the inherited word classes)::

            sage: m = WordMorphism('a->ab,b->a')
            sage: W = m.domain()
            sage: from sage.combinat.words.morphic import WordDatatype_morphic
            sage: w = WordDatatype_morphic(W, m, 'a')
            sage: type(w)
            <class 'sage.combinat.words.morphic.WordDatatype_morphic'>
            sage: w.representation(5)                                                   # needs sage.modules
            [1, 0, 0, 0]
        """
        letters_to_int = {a: i for i, a in enumerate(self._alphabet)}
        position = letters_to_int[self._letter]
        M = self._morphism.incidence_matrix()
        vMk = vector([1] * len(self._alphabet))
        length_of_images = []
        while vMk[position] <= n:
            length_of_images.append(vMk)
            vMk_next = vMk * M
            if vMk[position] == vMk_next[position]:
                raise IndexError(f'index (={n}) out of range, the fixed point is finite and has length {vMk[position]}')
            vMk = vMk_next
        k = len(length_of_images)
        letter_k = self._letter
        n_k = n
        path = []
        while k > 0:
            m_letter_k = self._morphism(letter_k)
            S = 0
            j = 0
            while S <= n_k:
                a = m_letter_k[j]
                i = letters_to_int[a]
                pile_length = length_of_images[k - 1][i]
                S += pile_length
                j += 1
            path.append(j - 1)
            n_k -= S - pile_length
            letter_k = a
            k -= 1
        return path

    def _func(self, key):
        """
        Return a letter of a fixed point of a morphism on position ``key``.

        INPUT:

        - ``self`` -- a fixed point of a morphism
        - ``key`` -- integer; the position

        OUTPUT: a letter

        EXAMPLES::

            sage: m = WordMorphism("a->ab,b->a")
            sage: w = m.fixed_point("a")
            sage: w[0]                                                                  # needs sage.modules
            'a'
            sage: w[5]                                                                  # needs sage.modules
            'a'
            sage: w[10000]                                                              # needs sage.modules
            'a'

        TESTS:

        Accessing this method from an instance of the current class
        (without using the inherited word classes)::

            sage: m = WordMorphism('a->ab,b->a')
            sage: W = m.domain()
            sage: from sage.combinat.words.morphic import WordDatatype_morphic
            sage: w = WordDatatype_morphic(W, m, 'a')
            sage: w._func(5)                                                            # needs sage.modules
            'a'
        """
        letter = self._letter
        for a in self.representation(key):
            letter = (self._morphism(letter))[a]
        return self._coding[letter]

    def __iter__(self) -> Iterator:
        r"""
        Return an iterator of the letters of the fixed point of ``self``
        starting with ``letter``.

        If w is the iterated word, then this iterator: outputs the elements
        of morphism[ w[i] ], appends morphism[ w[i+1] ] to w, increments i.

        INPUT:

        - ``self`` -- an endomorphism, must be prolongable on
          letter

        - ``letter`` -- a letter in the domain of ``self``

        OUTPUT: iterator of the fixed point

        EXAMPLES::

            sage: m = WordMorphism("a->ab,b->a")
            sage: w = m.fixed_point("a")
            sage: it = iter(w)
            sage: [next(it) for _ in range(10)]
            ['a', 'b', 'a', 'a', 'b', 'a', 'b', 'a', 'a', 'b']

        Works with erasing morphisms::

            sage: m = WordMorphism('a->abc,b->,c->')
            sage: w = m.fixed_point("a")
            sage: list(w)
            ['a', 'b', 'c']

        The morphism must be prolongable on the letter or the iterator will
        be empty::

            sage: list(m.fixed_point("b"))
            Traceback (most recent call last):
            ...
            TypeError: self must be prolongable on b

        The morphism must be an endomorphism::

            sage: m = WordMorphism('a->ac,b->aac')
            sage: w = m.fixed_point('a')
            Traceback (most recent call last):
            ...
            TypeError: self (=a->ac, b->aac) is not self-composable

        We check that :issue:`8595` is fixed::

            sage: s = WordMorphism({('a', 1):[('a', 1), ('a', 2)], ('a', 2):[('a', 1)]})
            sage: w = s.fixed_point(('a', 1))
            sage: it = iter(w)
            sage: next(it)
            ('a', 1)

        This shows that issue :issue:`13668` has been resolved::

            sage: s = WordMorphism({1:[1,2],2:[2,3],3:[4],4:[5],5:[6],6:[7],7:[8],8:[9],9:[10],10:[1]})
            sage: (s^7).fixed_points()
            [word: 1223234234523456234567234567823456789234...,
             word: 2,3,4,5,6,7,8,9,10,1,1,2,1,2,2,3,1,2,2,3,2,3,4,1,2,2,3,2,3,4,2,3,4,5,1,2,2,3,2,3,...]
            sage: (s^7).reversal().fixed_points()
            []
        """
        w = iter(self._morphism.image(self._letter))
        while True:
            try:
                for a in self._morphism.image(next(w)):
                    yield self._coding[a]
                next_w = next(w)
                w = chain([next_w], w, self._morphism.image(next_w))
            except StopIteration:
                return



# distutils: language = c++

from cysignals.signals cimport sig_on, sig_off
from memory_allocator cimport MemoryAllocator


def mcqd(G):
    """
    Compute the max clique using MCQD.

    INPUT:

    - ``G`` -- a graph

    TESTS::

        sage: from sage.graphs.mcqd import mcqd         # optional - mcqd
        sage: for i in range(10):                       # optional - mcqd
        ....:     g = graphs.RandomGNP(15,.5)
        ....:     if g.clique_number() != len(mcqd(g)):
        ....:         print("This is dead wrong !")
    """
    cdef int n = G.order()

    # - c0 is the adjacency matrix
    # - c points toward each row of the matrix
    # - qmax stores the max clique
    cdef MemoryAllocator mem = MemoryAllocator()
    cdef bool ** c = <bool **> mem.allocarray(n, sizeof(bool *))
    cdef bool * c0 = <bool *> mem.calloc(n*n, sizeof(bool))
    cdef int * qmax = <int *> mem.allocarray(n, sizeof(int))

    c[0] = c0

    # Defines c
    cdef int i, ui, vi
    for i in range(1, n):
        c[i] = c[i - 1] + n

    # Defines the adjacency matrix
    cdef list vertices = G.vertices(sort=False)
    cdef dict vertex_to_id = {v: i for i, v in enumerate(vertices)}

    for u in G:
        ui = vertex_to_id[u]
        for v in G.neighbors(u):
            vi = vertex_to_id[v]
            c[ui][vi] = 1
            c[vi][ui] = 1

    # Calls the solver
    cdef int clique_number
    cdef Maxclique * C = new Maxclique(c, n)
    sig_on()
    C.mcqdyn(qmax, clique_number)
    sig_off()

    # Returns the answer
    cdef list answer = [vertices[qmax[i]] for i in range(clique_number)]
    del C

    return answer



r"""
Pickling for the old CDF vector class

AUTHORS:

- Jason Grout

TESTS::

    sage: v = vector(CDF,[(1,-1), (2,pi), (3,5)]); v                                    # needs sage.symbolic
    (1.0 - 1.0*I, 2.0 + 3.141592653589793*I, 3.0 + 5.0*I)
    sage: loads(dumps(v)) == v                                                          # needs sage.symbolic
    True
"""

###############################################################################
#       Copyright (C) 2008 Jason Grout <jason-sage@creativetrax.com>
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
###############################################################################

from sage.modules.vector_complex_double_dense import Vector_complex_double_dense

ComplexDoubleVectorSpaceElement = Vector_complex_double_dense



# distutils: libraries = NTL_LIBRARIES gmp m
# distutils: extra_compile_args = NTL_CFLAGS
# distutils: include_dirs = NTL_INCDIR
# distutils: library_dirs = NTL_LIBDIR
# distutils: extra_link_args = NTL_LIBEXTRA
# distutils: language = c++
r"""
A ``PowComputer`` for relative extensions

This module provides helper classes for the various kinds of relative `p`-adic
extensions. You should never have to access these directly, unless you are
working on linkages or other low-level `p`-adics code within the Sage library.

AUTHORS:

- David Roe, Julian Rüth (2017-06-11): initial version
"""
# ****************************************************************************
#       Copyright (C) 2017 David Roe <roed.math@gmail.com>
#                     2017 Julian Rüth <julian.rueth@fsfe.org>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.rings.integer cimport Integer
from sage.misc.cachefunc import cached_method


cdef class PowComputer_relative(PowComputer_class):
    r"""
    Base class for a ``PowComputer`` for use in `p`-adics implemented by Sage
    Polynomials.

    For a description of inputs see :func:`PowComputer_relative_maker`.

    EXAMPLES::

        sage: from sage.rings.padics.pow_computer_relative import PowComputer_relative_maker
        sage: R.<a> = ZqFM(25)
        sage: S.<x> = R[]
        sage: f = x^3 - 5*x - 5*a
        sage: RFP = R.change(field=False, show_prec=False, type='floating-point')
        sage: shift_seed = (-f[:3] // 5).change_ring(RFP)
        sage: PC = PowComputer_relative_maker(3, 20, 20, 60, False, f, shift_seed, 'fixed-mod')

    TESTS::

        sage: from sage.rings.padics.pow_computer_relative import PowComputer_relative
        sage: isinstance(PC, PowComputer_relative)                                      # needs sage.libs.flint
        True
    """
    def __cinit__(self, Integer prime, long cache_limit, long prec_cap, long ram_prec_cap, bint in_field, poly, shift_seed):
        r"""
        TESTS::

            sage: from sage.rings.padics.pow_computer_relative import PowComputer_relative_maker
            sage: R.<a> = ZqFM(25)
            sage: S.<x> = R[]
            sage: f = x^3 - 5*x - 5*a
            sage: RFP = R.change(field=False, show_prec=False, type='floating-point')
            sage: shift_seed = (-f[:3] // 5).change_ring(RFP)
            sage: PC = PowComputer_relative_maker(3, 20, 20, 60, False, f, shift_seed, 'fixed-mod')
        """
        self._allocated = 4

    def __init__(self, Integer prime, long cache_limit, long prec_cap, long ram_prec_cap, bint in_field, poly, shift_seed):
        r"""
        TESTS::

            sage: from sage.rings.padics.pow_computer_relative import PowComputer_relative_maker
            sage: R.<a> = ZqFM(25)
            sage: S.<x> = R[]; f = x^3 - 5*x - 5*a
            sage: RFP = R.change(field=False, show_prec=False, type='floating-point')
            sage: shift_seed = (-f[:3] // 5).change_ring(RFP)
            sage: PC = PowComputer_relative_maker(5, 20, 20, 60, False, f, shift_seed, 'fixed-mod')  # indirect doctest
            sage: TestSuite(PC).run()
        """
        PowComputer_class.__init__(self, prime, cache_limit, prec_cap, ram_prec_cap, in_field, poly, shift_seed)
        self.e = poly.degree() * poly.base_ring().absolute_e()
        self.f = poly.base_ring().absolute_f()

        self.modulus = poly

        self.tmp_cconv_out = poly.parent()()
        self.tmp_ccoeffs = poly.parent()()
        self.tmp_ccmp_a = poly.parent()()
        self.tmp_ccmp_b = poly.parent()()
        self.shift_rem = poly.parent()()
        self.aliasing = poly.parent()()

        self.base_ring = poly.base_ring()
        self.poly_ring = poly.parent()
        self._shift_seed = shift_seed

    def __dealloc__(self):
        r"""
        TESTS::

            sage: from sage.rings.padics.pow_computer_relative import PowComputer_relative_maker
            sage: R.<a> = ZqFM(25)
            sage: S.<x> = R[]
            sage: f = x^3 - 5*x - 5*a
            sage: RFP = R.change(field=False, show_prec=False, type='floating-point')
            sage: shift_seed = (-f[:3] // 5).change_ring(RFP)
            sage: PC = PowComputer_relative_maker(5, 20, 20, 60, False, f, shift_seed, 'fixed-mod')
            sage: del PC
        """

    def __reduce__(self):
        r"""
        Return a picklable representation of this ``PowComputer``.

        TESTS::

            sage: from sage.rings.padics.pow_computer_relative import PowComputer_relative_maker
            sage: R.<a> = ZqFM(25)
            sage: S.<x> = R[]
            sage: f = x^3 - 5*x - 5*a
            sage: RFP = R.change(field=False, show_prec=False, type='floating-point')
            sage: shift_seed = (-f[:3] // 5).change_ring(RFP)
            sage: PC = PowComputer_relative_maker(5, 20, 20, 60, False, f, shift_seed, 'fixed-mod')  # indirect doctest
            sage: loads(dumps(PC)) == PC
            True
        """
        return PowComputer_relative_maker, (self.prime, self.cache_limit, self.prec_cap, self.ram_prec_cap, self.in_field, self.polynomial(), self._shift_seed, self._prec_type)

    def _repr_(self):
        r"""
        Return a string representation of this ``PowComputer``.

        EXAMPLES::

            sage: from sage.rings.padics.pow_computer_relative import PowComputer_relative_maker
            sage: R.<a> = ZqFM(25,print_pos=False,show_prec=False)
            sage: S.<x> = R[]
            sage: f = x^3 + 5*x + 5*a
            sage: RFP = R.change(field=False, show_prec=False, type='floating-point')
            sage: shift_seed = (-f[:3] // 5).change_ring(RFP)
            sage: PowComputer_relative_maker(5, 20, 20, 60, False, f, shift_seed, 'fixed-mod')  # indirect doctest
            Relative PowComputer for modulus x^3 + 5*x + a*5
        """
        return "Relative PowComputer for modulus %s" % (self.modulus,)

    cdef unsigned long capdiv(self, unsigned long n) noexcept:
        r"""
        Return `\lceil n/e \rceil`.
        """
        if self.e == 1:
            return n
        if n == 0:
            return 0
        return (n - 1)/self.e + 1

    def polynomial(self, n=None, var='x'):
        r"""
        Return the modulus of the `p`-adic extension that is handled by this
        ``PowComputer``.

        EXAMPLES::

            sage: from sage.rings.padics.pow_computer_relative import PowComputer_relative_maker
            sage: R.<a> = ZqFM(25)
            sage: S.<x> = R[]; f = x^3 - 5*x - 5*a
            sage: RFP = R.change(field=False, show_prec=False, type='floating-point')
            sage: shift_seed = (-f[:3] // 5).change_ring(RFP)
            sage: PC = PowComputer_relative_maker(5, 20, 20, 60, False, f, shift_seed, 'fixed-mod')  # indirect doctest
            sage: PC.polynomial() is f
            True
        """
        return self.modulus


cdef class PowComputer_relative_eis(PowComputer_relative):
    r"""
    A ``PowComputer`` for a relative extension defined by an Eisenstein polynomial.

    For a description of inputs see :func:`PowComputer_relative_maker`.

    EXAMPLES::

        sage: from sage.rings.padics.pow_computer_relative import PowComputer_relative_eis, PowComputer_relative_maker
        sage: R.<a> = ZqFM(25)
        sage: S.<x> = R[]; f = x^3 - 5*x - 5*a
        sage: RFP = R.change(field=False, show_prec=False, type='floating-point')
        sage: shift_seed = (-f[:3] // 5).change_ring(RFP)
        sage: PC = PowComputer_relative_maker(5, 20, 20, 60, False, f, shift_seed, 'fixed-mod')

    TESTS::

        sage: isinstance(PC, PowComputer_relative_eis)                                  # needs sage.libs.flint
        True
    """
    def __init__(self, Integer prime, long cache_limit, long prec_cap, long ram_prec_cap, bint in_field, poly, shift_seed):
        r"""
        TESTS::

            sage: from sage.rings.padics.pow_computer_relative import PowComputer_relative_maker
            sage: R.<a> = ZqFM(25)
            sage: S.<x> = R[]; f = x^3 - 5*x - 5*a
            sage: RFP = R.change(field=False, show_prec=False, type='floating-point')
            sage: shift_seed = (-f[:3] // 5).change_ring(RFP)
            sage: PC = PowComputer_relative_maker(5, 20, 20, 60, False, f, shift_seed, 'fixed-mod')
            sage: TestSuite(PC).run()
        """
        PowComputer_relative.__init__(self, prime, cache_limit, prec_cap, ram_prec_cap, in_field, poly, shift_seed)
        self._inv_shift_seed = self.invert(shift_seed, self.ram_prec_cap)

    cpdef Polynomial_generic_dense invert(self, Polynomial_generic_dense a, long prec):
        r"""
        Return the inverse of ``a``.

        INPUT:

        - ``a`` -- a `p`-adic element, represented as a reduced
          Polynomial in ``poly_ring``

        - ``prec`` -- a ``long``, the required precision

        OUTPUT: a polynomial ``b`` such that ``a*b`` is one modulo `π^\mathrm{prec}`

        EXAMPLES::

            sage: from sage.rings.padics.pow_computer_relative import PowComputer_relative_maker
            sage: R.<a> = ZqFM(25,3)
            sage: S.<x> = R[]; f = x^3 - 5*x - 5*a
            sage: W.<w> = R.ext(f)
            sage: g = 1 + 2*w; ginv = ~g
            sage: ginv
            1 + 3*w + 4*w^2 + 2*w^3 + (3*a + 3)*w^4 + ... + (3*a + 2)*w^8
            sage: RFP = R.change(field=False, show_prec=False, type='floating-point')
            sage: shift_seed = (-f[:3] // 5).change_ring(RFP)
            sage: PC = PowComputer_relative_maker(5, 3, 3, 9, False, f, shift_seed, 'fixed-mod')
            sage: g = 1 + 2*x
            sage: ginv = PC.invert(g, 5); ginv
            (4 + (3*a + 1)*5 + (2*a + 2)*5^2)*x^2 + (3 + (a + 1)*5 + (3*a + 2)*5^2)*x + 1 + 2*a*5 + 2*5^2
        """
        k = self.base_ring.residue_field()
        a0 = k(a[0])
        if a0.is_zero():
            raise ValueError("element has no inverse")
        K = self.base_ring.change(field=True)
        Qpmodulus = self.modulus.change_ring(K)
        Qpa = a.change_ring(K)
        R = Qpa.parent()
        inv = R([~a0])
        curprec = 1
        while curprec < prec:
            # Newton iteration
            inv = 2*inv - inv**2 * Qpa
            curprec *= 2
            inv = inv % Qpmodulus
        return inv.change_ring(self.base_ring)

    @cached_method
    def px_pow(self, r):
        r"""
        Return `p/π^r` where π is the uniformizer and `p` is the uniformizer of
        the base ring (not necessarily an integer.)

        INPUT:

        - ``r`` -- integer with 0 <= r < e

        OUTPUT: a reduced polynomial in π

        EXAMPLES::

            sage: R.<a> = Zq(25, prec=3)
            sage: S.<x> = R[]; f = x^3 - 5*x - 5*a
            sage: W.<w> = R.ext(f)
            sage: elt = W.prime_pow.px_pow(2); elt
            ((4*a + 4) + (4*a + 1)*5 + (4*a + 2)*5^2)*x^2 + ((2*a + 3) + (2*a + 4)*5 + (2*a + 4)*5^2)*x + (a + 1)*5 + 3*5^2 + 2*5^3
        """
        if r < 0:
            raise ValueError("r must be nonnegative")
        elif r == 0:
            return self.poly_ring(self.base_ring.uniformizer())
        elif r >= self.e:
            raise NotImplementedError
        else:
            return (self._inv_shift_seed << (self.e-r)) % self.modulus

    @cached_method
    def pxe_pow(self, r):
        r"""
        Return the ``r``-th power of the unit `p/π^e` where `e` is the relative
        ramification index and `p` is the uniformizer of the base ring (not necessarily an integer.)

        INPUT:

        - ``r`` -- nonnegative integer

        OUTPUT: a reduced polynomial in π

        EXAMPLES::

            sage: R.<a> = Zq(25, prec=3)
            sage: S.<x> = R[]; f = x^3 - 5*x - 5*a
            sage: W.<w> = R.ext(f)
            sage: elt = W.prime_pow.pxe_pow(2); elt
            ((4*a + 2) + (a + 4)*5 + 2*a*5^2)*x^2 + ((a + 2) + (a + 2)*5 + (2*a + 4)*5^2)*x + (a + 1) + (3*a + 2)*5 + (2*a + 2)*5^2
        """
        if r < 0:
            raise ValueError("r must be nonnegative")
        elif r == 0:
            return self.poly_ring.one()
        elif r == 1:
            return self._inv_shift_seed
        elif r % 2:
            return (self.pxe_pow(r-1) * self.pxe_pow(1)) % self.modulus
        else:
            return (self.pxe_pow(r//2)*self.pxe_pow(r//2)) % self.modulus

    @cached_method
    def uniformizer_pow(self, r):
        r"""
        Return the ``r``-th power of the uniformizer.

        INPUT:

        - ``r`` -- nonnegative integer

        OUTPUT: a reduced polynomial in π

        EXAMPLES::

            sage: R.<a> = Zq(25, prec=3)
            sage: S.<x> = R[]; f = x^3 - 5*x - 5*a
            sage: W.<w> = R.ext(f)
            sage: W.prime_pow.uniformizer_pow(2)
            x^2
        """
        if r < 0:
            raise ValueError("r must be nonnegative")
        elif r == 0:
            return self.poly_ring.one()
        elif r < self.e:
            return self.poly_ring.one() << r
        elif r % 2:
            return (self.uniformizer_pow(r-1) << 1) % self.modulus
        else:
            return (self.uniformizer_pow(r//2) * self.uniformizer_pow(r//2)) % self.modulus


def PowComputer_relative_maker(prime, cache_limit, prec_cap, ram_prec_cap, in_field, poly, shift_seed, prec_type):
    r"""
    Create a ``PowComputer``.

    INPUT:

    - ``prime`` -- a uniformizer in the base ring

    - ``cache_limit`` -- nonnegative integer, controlling the caching. The
      ``PowComputer`` caches frequently used things like powers of ``prime``
      This parameter, e.g., controls up to which power these are cached.

    - ``prec_cap`` -- the power of ``prime`` modulo which elements of largest
      precision are defined

    - ``ram_prec_cap`` -- approximately ``e*prec_cap``, where ``e`` is
      the relative ramification degree of the extension.  For a ramified
      extension this is what Sage calls the precision cap of the ring.  In
      fact, it's possible to have rings with precision cap not a multiple of
      `e`, in which case the actual relationship between ``ram_prec_cap`` and
      ``prec_cap`` is that ``prec_cap = ceil(n/e)``

    - ``in_field`` -- boolean; whether the associated ring is actually a
      field

    - ``poly`` -- the polynomial defining the extension

    - ``prec_type`` -- one of ``'capped-rel'``, ``'capped-abs'`` or
      ``'fixed-mod'``, ``'floating-point'``; the precision type of the ring

    .. NOTE::

        Because of the way templates work, this function imports the class of
        its return value from the appropriate element files.  This means that
        the returned PowComputer will have the appropriate compile-time-type
        for Cython.

    EXAMPLES::

        sage: from sage.rings.padics.pow_computer_relative import PowComputer_relative_maker
        sage: R.<a> = ZqFM(25, prec=2)
        sage: S.<x> = R[]
        sage: f = x^3 - 5*x - 5*a
        sage: W.<w> = R.extension(f)
        sage: PC = W.prime_pow  # indirect doctest
        sage: PC
        Relative PowComputer for modulus x^3 + (4*5 + 4*5^2)*x + 4*a*5 + 4*a*5^2
    """
    PC = PowComputer_relative_eis(prime, cache_limit, prec_cap, ram_prec_cap, in_field, poly, shift_seed)
    # We have to set this here because the signature of __cinit__ in PowComputer_base doesn't allow for prec_type to be passed.
    PC._prec_type = prec_type
    return PC



from sage.libs.flint.arb_poly cimport *
from sage.rings.polynomial.polynomial_element cimport Polynomial

cdef class Polynomial_real_arb(Polynomial):
    cdef arb_poly_struct[1] _poly # https://github.com/cython/cython/issues/1984
    cdef Polynomial_real_arb _new(self)



# sage.doctest: needs sage.symbolic
"""
Implicit plots
"""
from sage.plot.plot3d.implicit_surface import ImplicitSurface


def implicit_plot3d(f, xrange, yrange, zrange, **kwds):
    r"""
    Plot an isosurface of a function.

    INPUT:

    - ``f`` -- function

    - ``xrange`` -- a 2-tuple (x_min, x_max) or a 3-tuple (x, x_min, x_max)

    - ``yrange`` -- a 2-tuple (y_min, y_max) or a 3-tuple (y, y_min, y_max)

    - ``zrange`` -- a 2-tuple (z_min, z_max) or a 3-tuple (z, z_min, z_max)

    - ``plot_points`` -- (default: "automatic", which is 40) the number of
      function evaluations in each direction. (The number of cubes in the
      marching cubes algorithm will be one less than this). Can be a triple of
      integers, to specify a different resolution in each of x,y,z.

    - ``contour`` -- (default: 0) plot the isosurface ``f(x,y,z) == contour``.
      Can be a list, in which case multiple contours are plotted

    - ``region`` -- (default: ``None``) if region is given, it must be a Python
      callable. Only segments of the surface where ``region(x,y,z)`` returns a
      number `>0` will be included in the plot. (Note that returning a Python
      boolean is acceptable, since ``True == 1`` and ``False == 0``).

    EXAMPLES::

        sage: var('x,y,z')
        (x, y, z)

    A simple sphere::

        sage: implicit_plot3d(x^2+y^2+z^2==4, (x,-3,3), (y,-3,3), (z,-3,3))
        Graphics3d Object

    .. PLOT::

        var('x,y,z')
        F = x**2 + y**2 + z**2
        P = implicit_plot3d(F==4, (x,-3,3), (y,-3,3), (z,-3,3))
        sphinx_plot(P)

    A nested set of spheres with a hole cut out::

        sage: F = x^2 + y^2 + z^2
        sage: P = implicit_plot3d(F, (x,-2,2), (y,-2,2), (z,-2,2), plot_points=60,
        ....:                     contour=[1,3,5], color='aquamarine',
        ....:                     region=lambda x,y,z: x<=0.2 or y>=0.2 or z<=0.2)
        sage: P.show(viewer='tachyon')  # long time

    .. PLOT::

        var('x,y,z')
        F = x**2 + y**2 + z**2
        P = implicit_plot3d(F, (x,-2,2), (y,-2,2), (z,-2,2), plot_points=60, contour=[1,3,5],
                            region=lambda x,y,z: x<=0.2 or y>=0.2 or z<=0.2, color='aquamarine')
        sphinx_plot(P)

    A very pretty example, attributed to Douglas Summers-Stay (`archived page
    <http://web.archive.org/web/20080529033738/http://iat.ubalt.edu/summers/math/platsol.htm>`_)::

        sage: T = RDF(golden_ratio)
        sage: F = 2 - (cos(x+T*y) + cos(x-T*y) + cos(y+T*z)
        ....:           + cos(y-T*z) + cos(z-T*x) + cos(z+T*x))
        sage: r = 4.77
        sage: implicit_plot3d(F, (x,-r,r), (y,-r,r), (z,-r,r),
        ....:                 plot_points=40, color='darkkhaki').show(viewer='tachyon')

    .. PLOT::

        var('x,y,z')
        T = RDF(golden_ratio)
        F = 2 - (cos(x+T*y) + cos(x-T*y) + cos(y+T*z) + cos(y-T*z) + cos(z-T*x) + cos(z+T*x))
        r = 4.77
        V = implicit_plot3d(F, (x,-r,r), (y,-r,r), (z,-r,r), plot_points=40, color='darkkhaki')
        sphinx_plot(V)

    As I write this (but probably not as you read it), it's almost Valentine's
    day, so let's try a heart (from http://mathworld.wolfram.com/HeartSurface.html)

    ::

        sage: F = (x^2+9/4*y^2+z^2-1)^3 - x^2*z^3 - 9/(80)*y^2*z^3
        sage: r = 1.5
        sage: V = implicit_plot3d(F, (x,-r,r), (y,-r,r), (z,-r,r),
        ....:                     plot_points=80, color='red', smooth=False)
        sage: V.show(viewer='tachyon')  # long time

    .. PLOT::

        var('x,y,z')
        F = (x**2+9.0/4.0*y**2+z**2-1)**3 - x**2*z**3 - 9.0/(80)*y**2*z**3
        r = 1.5
        V = implicit_plot3d(F, (x,-r,r), (y,-r,r), (z,-r,r), plot_points=80, color='red', smooth=False)
        sphinx_plot(V)

    The same examples also work with the default Jmol viewer; for example::

        sage: T = RDF(golden_ratio)
        sage: F = 2 - (cos(x + T*y) + cos(x - T*y) + cos(y + T*z)
        ....:           + cos(y - T*z) + cos(z - T*x) + cos(z + T*x))
        sage: r = 4.77
        sage: implicit_plot3d(F, (x,-r,r), (y,-r,r), (z,-r,r),
        ....:                 plot_points=40, color='deepskyblue').show()

    Here we use ``smooth=True`` with a Tachyon graph::

        sage: implicit_plot3d(x^2 + y^2 + z^2, (x,-2,2), (y,-2,2), (z,-2,2),
        ....:                 contour=4, color='deepskyblue', smooth=True)
        Graphics3d Object

    .. PLOT::

        var('x,y,z')
        F = x**2 + y**2 + z**2
        P = implicit_plot3d(F, (x,-2,2), (y,-2,2), (z,-2,2), contour=4, color='deepskyblue', smooth=True)
        sphinx_plot(P)

    We explicitly specify a gradient function (in conjunction with ``smooth=True``)
    and invert the normals::

        sage: gx = lambda x, y, z: -(2*x + y^2 + z^2)
        sage: gy = lambda x, y, z: -(x^2 + 2*y + z^2)
        sage: gz = lambda x, y, z: -(x^2 + y^2 + 2*z)
        sage: implicit_plot3d(x^2+y^2+z^2, (x,-2,2), (y,-2,2), (z,-2,2),
        ....:                 contour=4, plot_points=40, smooth=True,
        ....:                 gradient=(gx, gy, gz)).show(viewer='tachyon')

    .. PLOT::

        var('x,y,z')
        gx = lambda x, y, z: -(2*x + y**2 + z**2)
        gy = lambda x, y, z: -(x**2 + 2*y + z**2)
        gz = lambda x, y, z: -(x**2 + y**2 + 2*z)
        P = implicit_plot3d(x**2+y**2+z**2, (x,-2,2), (y,-2,2), (z,-2,2), contour=4,
                           plot_points=40, smooth=True, gradient=(gx, gy, gz))
        sphinx_plot(P)

    A graph of two metaballs interacting with each other::

        sage: def metaball(x0, y0, z0):
        ....:     return 1 / ((x-x0)^2+(y-y0)^2+(z-z0)^2)
        sage: implicit_plot3d(metaball(-0.6,0,0) + metaball(0.6,0,0),
        ....:                 (x,-2,2), (y,-2,2), (z,-2,2),
        ....:                 plot_points=60, contour=2, color='seagreen')
        Graphics3d Object

    .. PLOT::

        var('x,y,z')
        def metaball(x0, y0, z0): return 1 / ((x-x0)**2+(y-y0)**2+(z-z0)**2)
        P = implicit_plot3d(metaball(-0.6,0,0) + metaball(0.6,0,0), (x,-2,2), (y,-2,2), (z,-2,2), plot_points=60, contour=2, color='seagreen')
        sphinx_plot(P)

    One can also color the surface using a coloring function and a
    colormap as follows. Note that the coloring function must take
    values in the interval `[0,1]`. ::

        sage: t = (sin(3*z)**2).function(x,y,z)
        sage: cm = colormaps.gist_rainbow
        sage: G = implicit_plot3d(x^2 + y^2 + z^2, (x,-2,2), (y,-2,2), (z,-2, 2),
        ....:                     contour=4, color=(t,cm), plot_points=100)
        sage: G.show(viewer='tachyon')  # long time

    .. PLOT::

        var('x,y,z')
        t = (sin(3*z)**2).function(x,y,z)
        cm = colormaps.gist_rainbow
        G = implicit_plot3d(x**2 + y**2 + z**2, (x,-2,2), (y,-2,2), (z,-2, 2),
                            contour=4, color=(t,cm), plot_points=60)
        sphinx_plot(G)

    Here is another colored example::

        sage: x, y, z = var('x,y,z')
        sage: t = (x).function(x,y,z)
        sage: cm = colormaps.PiYG
        sage: G = implicit_plot3d(x^4 + y^2 + z^2, (x,-2,2), (y,-2,2),(z,-2,2),
        ....:                     contour=4, color=(t,cm), plot_points=40)
        sage: G
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        t = (x).function(x,y,z)
        cm = colormaps.PiYG
        G = implicit_plot3d(x**4 + y**2 + z**2, (x,-2,2),
                           (y,-2,2),(z,-2,2), contour=4, color=(t,cm), plot_points=40)
        sphinx_plot(G)

    MANY MORE EXAMPLES:

    A kind of saddle::

        sage: implicit_plot3d(x^3 + y^2 - z^2, (x,-2,2), (y,-2,2), (z,-2,2),
        ....:                 plot_points=60, contour=0, color='lightcoral')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        G = implicit_plot3d(x**3 + y**2 - z**2, (x,-2,2), (y,-2,2), (z,-2,2), plot_points=60, contour=0, color='lightcoral')
        sphinx_plot(G)

    A smooth surface with six radial openings::

        sage: implicit_plot3d(-(cos(x) + cos(y) + cos(z)),
        ....:                 (x,-4,4), (y,-4,4), (z,-4,4), color='orchid')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        G = implicit_plot3d(-(cos(x) + cos(y) + cos(z)), (x,-4,4), (y,-4,4), (z,-4,4), color='orchid')
        sphinx_plot(G)

    A cube composed of eight conjoined blobs::

        sage: F = x^2 + y^2 + z^2 + cos(4*x) + cos(4*y) + cos(4*z) - 0.2
        sage: implicit_plot3d(F, (x,-2,2), (y,-2,2), (z,-2,2), color='mediumspringgreen')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        F = x**2 + y**2 + z**2 + cos(4*x) + cos(4*y) + cos(4*z) - 0.2
        G = implicit_plot3d(F, (x,-2,2), (y,-2,2), (z,-2,2), color='mediumspringgreen')
        sphinx_plot(G)

    A variation of the blob cube featuring heterogeneously sized blobs::

        sage: F = x^2 + y^2 + z^2 + sin(4*x) + sin(4*y) + sin(4*z) - 1
        sage: implicit_plot3d(F, (x,-2,2), (y,-2,2), (z,-2,2), color='lavenderblush')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        F = x**2 + y**2 + z**2 + sin(4*x) + sin(4*y) + sin(4*z) - 1
        G = implicit_plot3d(F, (x,-2,2), (y,-2,2), (z,-2,2), color='lavenderblush')
        sphinx_plot(G)

    A Klein bottle::

        sage: G = x^2 + y^2 + z^2
        sage: F = (G+2*y-1)*((G-2*y-1)^2-8*z^2) + 16*x*z*(G-2*y-1)
        sage: implicit_plot3d(F, (x,-3,3), (y,-3.1,3.1), (z,-4,4), color='moccasin')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        G = x**2 + y**2 + z**2
        F = (G+2*y-1)*((G-2*y-1)**2-8*z**2)+16*x*z*(G-2*y-1)
        G = implicit_plot3d(F, (x,-3,3), (y,-3.1,3.1), (z,-4,4), color='moccasin')
        sphinx_plot(G)

    A lemniscate::

        sage: F = 4*x^2*(x^2+y^2+z^2+z) + y^2*(y^2+z^2-1)
        sage: implicit_plot3d(F, (x,-0.5,0.5), (y,-1,1), (z,-1,1), color='deeppink')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        F = 4*x**2*(x**2+y**2+z**2+z) + y**2*(y**2+z**2-1)
        G = implicit_plot3d(F, (x,-0.5,0.5), (y,-1,1), (z,-1,1), color='deeppink')
        sphinx_plot(G)

    Drope::

        sage: implicit_plot3d(z - 4*x*exp(-x^2-y^2), (x,-2,2), (y,-2,2), (z,-1.7,1.7),
        ....:                 color='darkcyan')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        G = implicit_plot3d(z - 4*x*exp(-x**2-y**2), (x,-2,2), (y,-2,2), (z,-1.7,1.7), color='darkcyan')
        sphinx_plot(G)

    A cube with a circular aperture on each face::

        sage: F = ((1/2.3)^2 * (x^2 + y^2 + z^2))^(-6) + ((1/2)^8 * (x^8 + y^8 + z^8))^6 - 1
        sage: implicit_plot3d(F, (x,-2,2), (y,-2,2), (z,-2,2), color='palevioletred')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        F = ((1/2.3)**2 * (x**2 + y**2 + z**2))**(-6) + ((1/2)**8 * (x**8 + y**8 + z**8))**6 - 1
        G = implicit_plot3d(F, (x,-2,2), (y,-2,2), (z,-2,2), color='palevioletred')
        sphinx_plot(G)

    A simple hyperbolic surface::

        sage: implicit_plot3d(x^2 + y - z^2, (x,-1,1), (y,-1,1), (z,-1,1),
        ....:                 color='darkslategray')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        G = implicit_plot3d(x**2 + y - z**2, (x,-1,1), (y,-1,1), (z,-1,1), color='darkslategray')
        sphinx_plot(G)

    A hyperboloid::

        sage: implicit_plot3d(x^2 + y^2 - z^2 -0.3, (x,-2,2), (y,-2,2), (z,-1.8,1.8),
        ....:                 color='honeydew')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        G = implicit_plot3d(x**2 + y**2 - z**2 -0.3, (x,-2,2), (y,-2,2), (z,-1.8,1.8), color='honeydew')
        sphinx_plot(G)

    Dupin cyclide (:wikipedia:`Dupin_cyclide`) ::

        sage: x, y, z, a, b, c, d = var('x,y,z,a,b,c,d')
        sage: a = 3.5
        sage: b = 3
        sage: c = sqrt(a^2 - b^2)
        sage: d = 2
        sage: F = (x^2 + y^2 + z^2 + b^2 - d^2)^2 - 4*(a*x-c*d)^2 - 4*b^2*y^2
        sage: implicit_plot3d(F, (x,-6,6), (y,-6,6), (z,-6,6), color='seashell')
        Graphics3d Object

    .. PLOT::

        x, y, z , a, b, c, d = var('x,y,z,a,b,c,d')
        a = 3.5
        b = 3
        c = sqrt(a**2 - b**2)
        d = 2
        F = (x**2 + y**2 + z**2 + b**2 - d**2)**2 - 4*(a*x-c*d)**2 - 4*b**2*y**2
        G = implicit_plot3d(F, (x,-6,6), (y,-6,6), (z,-6,6), color='seashell')
        sphinx_plot(G)

    Sinus::

        sage: implicit_plot3d(sin(pi*((x)^2+(y)^2))/2 + z, (x,-1,1), (y,-1,1), (z,-1,1),
        ....:                 color='rosybrown')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        G = implicit_plot3d(sin(pi*((x)**2+(y)**2))/2 + z, (x,-1,1), (y,-1,1), (z,-1,1), color='rosybrown')
        sphinx_plot(G)

    A torus::

        sage: implicit_plot3d((sqrt(x*x+y*y)-3)^2 + z*z - 1, (x,-4,4), (y,-4,4), (z,-1,1),
        ....:                 color='indigo')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        G = implicit_plot3d((sqrt(x*x+y*y)-3)**2 + z*z - 1, (x,-4,4), (y,-4,4), (z,-1,1), color='indigo')
        sphinx_plot(G)

    An octahedron::

        sage: implicit_plot3d(abs(x) + abs(y) + abs(z) - 1, (x,-1,1), (y,-1,1), (z,-1,1),
        ....:                 color='olive')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        G = implicit_plot3d(abs(x) + abs(y) + abs(z) - 1, (x,-1,1), (y,-1,1), (z,-1,1), color='olive')
        sphinx_plot(G)

    A cube::

        sage: implicit_plot3d(x^100 + y^100 + z^100 - 1, (x,-2,2), (y,-2,2), (z,-2,2),
        ....:                 color='lightseagreen')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        G = implicit_plot3d(x**100 + y**100 + z**100 - 1, (x,-2,2), (y,-2,2), (z,-2,2), color='lightseagreen')
        sphinx_plot(G)

    Toupie::

        sage: implicit_plot3d((sqrt(x*x+y*y)-3)^3 + z*z - 1, (x,-4,4), (y,-4,4), (z,-6,6),
        ....:                 color='mintcream')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        G = implicit_plot3d((sqrt(x*x+y*y)-3)**3 + z*z - 1, (x,-4,4), (y,-4,4), (z,-6,6), color='mintcream')
        sphinx_plot(G)

    A cube with rounded edges::

        sage: F = x^4 + y^4 + z^4 - (x^2 + y^2 + z^2)
        sage: implicit_plot3d(F, (x,-2,2), (y,-2,2), (z,-2,2), color='mediumvioletred')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        F = x**4 + y**4 + z**4 - (x**2 + y**2 + z**2)
        G = implicit_plot3d(F, (x,-2,2), (y,-2,2), (z,-2,2), color='mediumvioletred')
        sphinx_plot(G)

    Chmutov::

        sage: F = x^4 + y^4 + z^4 - (x^2 + y^2 + z^2 - 0.3)
        sage: implicit_plot3d(F, (x,-1.5,1.5), (y,-1.5,1.5), (z,-1.5,1.5),
        ....:                 color='lightskyblue')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        F = x**4 + y**4 + z**4 - (x**2 + y**2 + z**2 - 0.3)
        G = implicit_plot3d(F, (x,-1.5,1.5), (y,-1.5,1.5), (z,-1.5,1.5), color='lightskyblue')
        sphinx_plot(G)

    Further Chmutov::

        sage: F = 2*(x^2*(3-4*x^2)^2+y^2*(3-4*y^2)^2+z^2*(3-4*z^2)^2) - 3
        sage: implicit_plot3d(F, (x,-1.3,1.3), (y,-1.3,1.3), (z,-1.3,1.3),
        ....:                 color='darksalmon')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        F = 2*(x**2*(3-4*x**2)**2+y**2*(3-4*y**2)**2+z**2*(3-4*z**2)**2) - 3
        G = implicit_plot3d(F, (x,-1.3,1.3), (y,-1.3,1.3), (z,-1.3,1.3), color='darksalmon')
        sphinx_plot(G)

    Clebsch surface::

        sage: F_1 = 81 * (x^3+y^3+z^3)
        sage: F_2 = 189 * (x^2*(y+z)+y^2*(x+z)+z^2*(x+y))
        sage: F_3 = 54 * x * y * z
        sage: F_4 = 126 * (x*y+x*z+y*z)
        sage: F_5 = 9 * (x^2+y^2+z^2)
        sage: F_6 = 9 * (x+y+z)
        sage: F = F_1 - F_2 + F_3 + F_4 - F_5 + F_6 + 1
        sage: implicit_plot3d(F, (x,-1,1), (y,-1,1), (z,-1,1), color='yellowgreen')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        F_1 = 81 * (x**3+y**3+z**3)
        F_2 = 189 * (x**2*(y+z)+y**2*(x+z)+z**2*(x+y))
        F_3 = 54 * x * y * z
        F_4 = 126 * (x*y+x*z+y*z)
        F_5 = 9 * (x**2+y**2+z**2)
        F_6 = 9 * (x+y+z)
        F = F_1 - F_2 + F_3 + F_4 - F_5 + F_6 + 1
        G = implicit_plot3d(F, (x,-1,1), (y,-1,1), (z,-1,1), color='yellowgreen')
        sphinx_plot(G)

    Looks like a water droplet::

        sage: implicit_plot3d(x^2 + y^2 - (1-z)*z^2, (x,-1.5,1.5), (y,-1.5,1.5), (z,-1,1),
        ....:                 color='bisque')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        G = implicit_plot3d(x**2 +y**2 -(1-z)*z**2, (x,-1.5,1.5), (y,-1.5,1.5), (z,-1,1), color='bisque')
        sphinx_plot(G)

    Sphere in a cage::

        sage: F = (x^8+z^30+y^8-(x^4 + z^50 + y^4 -0.3)) * (x^2+y^2+z^2-0.5)
        sage: implicit_plot3d(F, (x,-1.2,1.2), (y,-1.3,1.3), (z,-1.5,1.5),
        ....:                 color='firebrick')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        F = (x**8+z**30+y**8-(x**4 + z**50 + y**4 -0.3)) * (x**2+y**2+z**2-0.5)
        G = implicit_plot3d(F, (x,-1.2,1.2), (y,-1.3,1.3), (z,-1.5,1.5), color='firebrick')
        sphinx_plot(G)

    Ortho circle::

        sage: F = (((x^2+y^2-1)^2+z^2) * ((y^2+z^2-1)^2+x^2) * ((z^2+x^2-1)^2+y^2)
        ....:      - 0.075^2 * (1+3*(x^2+y^2+z^2)))
        sage: implicit_plot3d(F, (x,-1.5,1.5), (y,-1.5,1.5), (z,-1.5,1.5),
        ....:                 color='lemonchiffon')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        F = ((x**2+y**2-1)**2+z**2) * ((y**2+z**2-1)**2+x**2) * ((z**2+x**2-1)**2+y**2)-0.075**2 * (1+3*(x**2+y**2+z**2))
        G = implicit_plot3d(F, (x,-1.5,1.5), (y,-1.5,1.5), (z,-1.5,1.5), color='lemonchiffon')
        sphinx_plot(G)

    Cube sphere::

        sage: F = 12 - ((1/2.3)^2 *(x^2 + y^2 + z^2))^-6 - ((1/2)^8 * (x^8 + y^8 + z^8))^6
        sage: implicit_plot3d(F, (x,-2,2), (y,-2,2), (z,-2,2), color='rosybrown')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        F = 12 - ((1/2.3)**2 *(x**2 + y**2 + z**2))**-6 - ( (1/2)**8 * (x**8 + y**8 + z**8) )**6
        G = implicit_plot3d(F, (x,-2,2), (y,-2,2), (z,-2,2), color='rosybrown')
        sphinx_plot(G)

    Two cylinders intersect to make a cross::

        sage: implicit_plot3d((x^2+y^2-1) * (x^2+z^2-1) - 1, (x,-3,3), (y,-3,3), (z,-3,3),
        ....:                 color='burlywood')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        G = implicit_plot3d((x**2+y**2-1) * (x**2+z**2-1) - 1, (x,-3,3), (y,-3,3), (z,-3,3), color='burlywood')
        sphinx_plot(G)

    Three cylinders intersect in a similar fashion::

        sage: implicit_plot3d((x^2+y^2-1) * (x^2+z^2-1) * (y^2+z^2-1) - 1,
        ....:                 (x,-3,3), (y,-3,3), (z,-3,3), color='aqua')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        G = implicit_plot3d((x**2+y**2-1) * (x**2+z**2-1) * (y**2+z**2-1)-1, (x,-3,3), (y,-3,3), (z,-3,3), color='aqua')
        sphinx_plot(G)

    A sphere-ish object with twelve holes, four on each XYZ plane::

        sage: implicit_plot3d(3*(cos(x)+cos(y)+cos(z)) + 4*cos(x)*cos(y)*cos(z),
        ....:                 (x,-3,3), (y,-3,3), (z,-3,3), color='orangered')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        G = implicit_plot3d(3*(cos(x)+cos(y)+cos(z)) + 4*cos(x)*cos(y)*cos(z), (x,-3,3), (y,-3,3), (z,-3,3), color='orangered')
        sphinx_plot(G)

    A gyroid::

        sage: implicit_plot3d(cos(x)*sin(y) + cos(y)*sin(z) + cos(z)*sin(x),
        ....:                 (x,-4,4), (y,-4,4), (z,-4,4), color='sandybrown')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        G = implicit_plot3d(cos(x)*sin(y) + cos(y)*sin(z) + cos(z)*sin(x), (x,-4,4), (y,-4,4), (z,-4,4), color='sandybrown')
        sphinx_plot(G)

    Tetrahedra::

        sage: implicit_plot3d((x^2+y^2+z^2)^2 + 8*x*y*z - 10*(x^2+y^2+z^2) + 25,
        ....:                 (x,-4,4), (y,-4,4), (z,-4,4), color='plum')
        Graphics3d Object

    .. PLOT::

        x, y, z = var('x,y,z')
        G = implicit_plot3d((x**2+y**2+z**2)**2 + 8*x*y*z - 10*(x**2+y**2+z**2) + 25, (x,-4,4), (y,-4,4), (z,-4,4), color='plum')
        sphinx_plot(G)

    TESTS:

    Test a separate resolution in the X direction; this should look like a
    regular sphere::

        sage: implicit_plot3d(x^2 + y^2 + z^2, (x,-2,2), (y,-2,2), (z,-2,2), plot_points=(10,40,40), contour=4)
        Graphics3d Object

    Test using different plot ranges in the different directions; each
    of these should generate half of a sphere.  Note that we need to use
    the ``aspect_ratio`` keyword to make it look right with the unequal
    plot ranges::

        sage: implicit_plot3d(x^2 + y^2 + z^2, (x,0,2), (y,-2,2), (z,-2,2), contour=4, aspect_ratio=1)
        Graphics3d Object

        sage: implicit_plot3d(x^2 + y^2 + z^2, (x,-2,2), (y,0,2), (z,-2,2), contour=4, aspect_ratio=1)
        Graphics3d Object

        sage: implicit_plot3d(x^2 + y^2 + z^2, (x,-2,2), (y,-2,2), (z,0,2), contour=4, aspect_ratio=1)
        Graphics3d Object

    Extra keyword arguments will be passed to show()::

        sage: implicit_plot3d(x^2 + y^2 + z^2, (x,-2,2), (y,-2,2), (z,-2,2), contour=4, viewer='tachyon')
        Graphics3d Object

    An implicit plot that does not include any surface in the view volume
    produces an empty plot::

        sage: implicit_plot3d(x^2 + y^2 + z^2 - 5000, (x,-2,2), (y,-2,2), (z,-2,2), plot_points=6)
        Graphics3d Object

    Make sure that implicit_plot3d does not error if the function cannot
    be symbolically differentiated::

        sage: implicit_plot3d(max_symbolic(x, y^2) - z, (x,-2,2), (y,-2,2), (z,-2,2), plot_points=6)
        Graphics3d Object

    TESTS:

    Check for :issue:`10599`::

        sage: var('x,y,z')
        (x, y, z)
        sage: M = matrix(3,[1,-1,-1,-1,3,1,-1,1,3])
        sage: v = 1/M.eigenvalues()[1]
        sage: implicit_plot3d(x^2+y^2+z^2==v, [x,-3,3], [y,-3,3],[z,-3,3])
        Graphics3d Object
    """
    # These options, related to rendering with smooth shading, are irrelevant
    # since IndexFaceSet does not support surface normals:
    # smooth: (default: ``False``) Whether to use vertex normals to produce a
    #   smooth-looking surface. False is slightly faster.
    # gradient: (default: None) If smooth is True (the default), then
    #   Tachyon rendering needs vertex normals. In that case, if gradient is None
    #   (the default), then we try to differentiate the function to get the
    #   gradient. If that fails, then we use central differencing on the scalar
    #   field. But it's also possible to specify the gradient; this must be either
    #   a single python callable that takes (x,y,z) and returns a tuple (dx,dy,dz)
    #   or a tuple of three callables that each take (x,y,z) and return dx, dy, dz
    #   respectively.

    G = ImplicitSurface(f, xrange, yrange, zrange, **kwds)
    G._set_extra_kwds(kwds)
    return G



"""
Pickling for the old RDF vector class

AUTHORS:

- Jason Grout

TESTS::

    sage: v = vector(RDF, [1,2,3,4])
    sage: loads(dumps(v)) == v
    True
"""
###############################################################################
#       Copyright (C) 2008 Jason Grout <jason-sage@creativetrax.com>
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
###############################################################################
from sage.modules.vector_real_double_dense import Vector_real_double_dense

RealDoubleVectorSpaceElement = Vector_real_double_dense



"""
Space of pseudomorphisms of free modules

AUTHORS:

- Xavier Caruso, Yossef Musleh (2024-09): initial version
"""
# ****************************************************************************
#  Copyright (C) 2024 Xavier Caruso <xavier.caruso@normalesup.org>
#                     Yossef Musleh <specialholonomy@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty
#    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License for more details; the full text
#  is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.structure.unique_representation import UniqueRepresentation

from sage.categories.homset import HomsetWithBase
from sage.matrix.matrix_space import MatrixSpace
from sage.structure.sequence import Sequence
from sage.rings.polynomial.ore_polynomial_ring import OrePolynomialRing
from sage.modules.free_module_pseudomorphism import FreeModulePseudoMorphism


class FreeModulePseudoHomspace(UniqueRepresentation, HomsetWithBase):
    r"""
    This class implements the space of pseudomorphisms with a fixed twist.

    For free modules, the elements of a pseudomorphism correspond to matrices
    which define the mapping on elements of a basis.

    This class is not supposed to be instantiated directly; the user should
    use instead the method :meth:`sage.rings.module.free_module.FreeModule_generic.pseudoHom`
    to create a space of pseudomorphisms.

    TESTS::

        sage: F = GF(125)
        sage: M = F^2
        sage: Frob = F.frobenius_endomorphism()
        sage: PHS = M.pseudoHom(Frob)
        sage: h = PHS([[1, 2], [1, 1]])
        sage: e = M((4*F.gen()^2 + F.gen() + 2, 4*F.gen()^2 + 4*F.gen() + 4))
        sage: h(e)
        (z3, 2*z3^2 + 3*z3 + 3)
    """
    Element = FreeModulePseudoMorphism

    @staticmethod
    def __classcall_private__(cls, domain, codomain, twist):
        r"""
        Constructs the space of pseudomorphisms with a given twist.

        INPUT:

        - ``domain`` -- a free module,  the domain of this pseudomorphism

        - ``codomain`` -- a free module, the codomain of this pseudomorphism

        - ``twist`` -- a twisting morphism/derivation or the corresponding
          Ore polynomial ring

        TESTS::

            sage: F = GF(125)
            sage: Frob = F.frobenius_endomorphism()
            sage: M = F^2
            sage: H = M.pseudoHom(Frob)
            sage: type(H)
            <class 'sage.modules.free_module_pseudohomspace.FreeModulePseudoHomspace_with_category'>

            sage: TestSuite(H).run()
        """
        ring = domain.base_ring()
        if codomain.base_ring() is not ring:
            raise ValueError("the domain and the codomain must be defined over the same ring")
        if isinstance(twist, OrePolynomialRing):
            ore = twist
            if ore.base_ring() is not ring:
                raise ValueError("base rings do not match")
        else:
            ore = OrePolynomialRing(ring, twist, names='x', polcast=False)
        if ore._derivation is not None:
            if not codomain.has_coerce_map_from(domain):
                raise ValueError("the domain does not coerce into the codomain")
        return cls.__classcall__(cls, domain, codomain, ore)

    def __init__(self, domain, codomain, ore):
        r"""
        Initialize this pseudohom space.

        INPUT:

        - ``domain`` -- a free module,  the domain of this pseudomorphism

        - ``codomain`` -- a free module, the codomain of this pseudomorphism

        - ``ore`` -- the underlying Ore polynomial ring (built from the
          twisting morphism and derivation)

        TESTS::

            sage: F = GF(125)
            sage: Frob = F.frobenius_endomorphism()
            sage: M = F^2
            sage: M.pseudoHom(Frob)
            Set of Pseudoendomorphisms (twisted by z3 |--> z3^5) of
            Vector space of dimension 2 over Finite Field in z3 of size 5^3
        """
        self._domain = domain
        self._codomain = codomain
        super().__init__(domain, codomain, category=None)
        self._ore = ore
        if isinstance(ore, OrePolynomialRing):
            self._morphism = ore._morphism
            self._derivation = ore._derivation
        else:
            self._morphism = self._derivation = None
        ring = ore.base_ring()
        self._matrix_space = MatrixSpace(ring, domain.dimension(), codomain.dimension())

    def _element_constructor_(self, f, side="left"):
        r"""
        Return the element of this parent constructed from the
        given data.

        TESTS::

            sage: F.<z> = GF(5^3)
            sage: Frob = F.frobenius_endomorphism()
            sage: V = F^2
            sage: H = V.pseudoHom(Frob)

            sage: H([[1, z], [z, z^2]])
            Free module pseudomorphism (twisted by z |--> z^5) defined by the matrix
            [  1   z]
            [  z z^2]
            Domain: Vector space of dimension 2 over Finite Field in z of size 5^3
            Codomain: Vector space of dimension 2 over Finite Field in z of size 5^3
        """
        return self.element_class(self, f, side)

    def __reduce__(self):
        r"""
        TESTS::

            sage: F = GF(125)
            sage: Frob = F.frobenius_endomorphism()
            sage: M = F^2
            sage: H = M.pseudoHom(Frob)
            sage: loads(dumps(M)) is M
            True
        """
        if self._derivation is None:
            twist = self._morphism
        else:
            twist = self._derivation
        return FreeModulePseudoHomspace, (self.domain(), self.codomain(), twist)

    def _repr_(self):
        r"""
        Returns a string representation of this pseudomorphism space.

        EXAMPLES::

            sage: Fq = GF(7^3)
            sage: Frob = Fq.frobenius_endomorphism()
            sage: V = Fq^2
            sage: V.pseudoHom(Frob)  # indirect doctest
            Set of Pseudoendomorphisms (twisted by z3 |--> z3^7) of
            Vector space of dimension 2 over Finite Field in z3 of size 7^3

        ::

            sage: V.pseudoHom(Frob, codomain=Fq^3)  # indirect doctest
            Set of Pseudomorphism (twisted by z3 |--> z3^7)
            from Vector space of dimension 2 over Finite Field in z3 of size 7^3
            to Vector space of dimension 3 over Finite Field in z3 of size 7^3

        ::

            sage: A.<t> = QQ[]
            sage: d = A.derivation()
            sage: M = A^3
            sage: M.pseudoHom(d)
            Set of Pseudoendomorphisms (twisted by d/dt) of Ambient free module of rank 3 over
            the principal ideal domain Univariate Polynomial Ring in t over Rational Field
        """
        twist = self._ore._repr_twist()
        if self.domain() is self.codomain():
            return "Set of Pseudoendomorphisms (%s) of %s" % (twist, self.domain())
        return "Set of Pseudomorphism (%s) from %s to %s" % (twist, self.domain(), self.codomain())

    def ore_ring(self, var='x'):
        r"""
        Return the underlying Ore polynomial ring, that is
        the Ore polynomial ring over the base field twisted
        by the twisting morphism and the twisting derivation
        attached to this homspace.

        INPUT:

        - ``var`` -- string (default: ``x``) the name of
          the variable

        EXAMPLES::

            sage: Fq.<z> = GF(7^3)
            sage: Frob = Fq.frobenius_endomorphism()
            sage: V = Fq^2
            sage: H = V.pseudoHom(Frob)

            sage: H.ore_ring()
            Ore Polynomial Ring in x over Finite Field in z of size 7^3 twisted by z |--> z^7

            sage: H.ore_ring('y')
            Ore Polynomial Ring in y over Finite Field in z of size 7^3 twisted by z |--> z^7
        """
        return self._ore.change_var(var)

    def matrix_space(self):
        r"""
        Return the matrix space used for representing the
        pseudomorphisms in this space.

        EXAMPLES::

            sage: Fq.<z> = GF(7^3)
            sage: Frob = Fq.frobenius_endomorphism()
            sage: V = Fq^2
            sage: W = Fq^3
            sage: H = V.pseudoHom(Frob, codomain=W)
            sage: H.matrix_space()
            Full MatrixSpace of 2 by 3 dense matrices over Finite Field in z of size 7^3
        """
        return self._matrix_space

    def basis(self, side="left"):
        r"""
        Return a basis for the underlying matrix space.

        The result does not depend on the `side` of the homspace, i.e.
        if matrices are acted upon on the left or on the right.

        EXAMPLES::

            sage: Fq = GF(7^3)
            sage: Frob = Fq.frobenius_endomorphism()
            sage: V = Fq^2
            sage: PHS = V.pseudoHom(Frob)
            sage: PHS.basis()
            [Free module pseudomorphism (twisted by z3 |--> z3^7) defined by the matrix
            [1 0]
            [0 0]
            Domain: Vector space of dimension 2 over Finite Field in z3 of size 7^3
            Codomain: Vector space of dimension 2 over Finite Field in z3 of size 7^3,
            Free module pseudomorphism (twisted by z3 |--> z3^7) defined by the matrix
            [0 1]
            [0 0]
            Domain: Vector space of dimension 2 over Finite Field in z3 of size 7^3
            Codomain: Vector space of dimension 2 over Finite Field in z3 of size 7^3,
            Free module pseudomorphism (twisted by z3 |--> z3^7) defined by the matrix
            [0 0]
            [1 0]
            Domain: Vector space of dimension 2 over Finite Field in z3 of size 7^3
            Codomain: Vector space of dimension 2 over Finite Field in z3 of size 7^3,
            Free module pseudomorphism (twisted by z3 |--> z3^7) defined by the matrix
            [0 0]
            [0 1]
            Domain: Vector space of dimension 2 over Finite Field in z3 of size 7^3
            Codomain: Vector space of dimension 2 over Finite Field in z3 of size 7^3]
        """
        return Sequence(self(mat) for mat in self._matrix_space.basis())

    def _test_additive_associativity(self, tester):
        r"""
        Test associativity for (not necessarily all) elements in this parent.

        This test is not relevant for pseudomorphisms because they are not
        stable by addition.

        TESTS::

            sage: Fq = GF(7^3)
            sage: Frob = Fq.frobenius_endomorphism()
            sage: V = Fq^2
            sage: PHS = V.pseudoHom(Frob)
            sage: TestSuite(PHS).run()  # indirect doctest
        """
        pass

    def _test_distributivity(self, tester):
        r"""
        Test distributivity for (not necessarily all) elements in this parent.

        This test is not relevant for pseudomorphisms because they are not
        stable by addition.

        TESTS::

            sage: Fq = GF(7^3)
            sage: Frob = Fq.frobenius_endomorphism()
            sage: V = Fq^2
            sage: PHS = V.pseudoHom(Frob)
            sage: TestSuite(PHS).run()  # indirect doctest
        """
        pass

    def _test_one(self, tester):
        r"""
        Test properties the identity element.

        This test is not relevant for pseudomorphisms because the identity
        is not a pseudomorphism in general.

        TESTS::

            sage: Fq = GF(7^3)
            sage: Frob = Fq.frobenius_endomorphism()
            sage: V = Fq^2
            sage: PHS = V.pseudoHom(Frob)
            sage: TestSuite(PHS).run()  # indirect doctest
        """
        pass

    def _test_zero(self, tester):
        r"""
        Test properties of the zero element.

        This test is not relevant for pseudomorphisms because the zero
        map is not a pseudomorphism in general.

        TESTS::

            sage: Fq = GF(7^3)
            sage: Frob = Fq.frobenius_endomorphism()
            sage: V = Fq^2
            sage: PHS = V.pseudoHom(Frob)
            sage: TestSuite(PHS).run()  # indirect doctest
        """
        pass
